@@ -1,17 +1,17 @@
-/* 
+/*
    Copyright (C) Andrew Tridgell 1996
    Copyright (C) Paul Mackerras 1996
-   
+
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
-   
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-   
+
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
@@ -23,7 +23,7 @@ int csum_length=2; /* initial value */
 
 #define CSUM_CHUNK 64
 
-int checksum_seed = 0;
+extern int checksum_seed;
 extern int protocol_version;
 
 /*
@@ -38,9 +38,9 @@ uint32 get_checksum1(char *buf1,int len)
 
     s1 = s2 = 0;
     for (i = 0; i < (len-4); i+=4) {
-	s2 += 4*(s1 + buf[i]) + 3*buf[i+1] + 2*buf[i+2] + buf[i+3] + 
+	s2 += 4*(s1 + buf[i]) + 3*buf[i+1] + 2*buf[i+2] + buf[i+3] +
 	  10*CHAR_OFFSET;
-	s1 += (buf[i+0] + buf[i+1] + buf[i+2] + buf[i+3] + 4*CHAR_OFFSET); 
+	s1 += (buf[i+0] + buf[i+1] + buf[i+2] + buf[i+3] + 4*CHAR_OFFSET);
     }
     for (; i < len; i++) {
 	s1 += (buf[i]+CHAR_OFFSET); s2 += s1;
@@ -62,15 +62,15 @@ void get_checksum2(char *buf,int len,char *sum)
 		len1 = len;
 		if (!buf1) out_of_memory("get_checksum2");
 	}
-	
+
 	mdfour_begin(&m);
-	
+
 	memcpy(buf1,buf,len);
 	if (checksum_seed) {
 		SIVAL(buf1,len,checksum_seed);
 		len += 4;
 	}
-	
+
 	for(i = 0; i + CSUM_CHUNK <= len; i += CSUM_CHUNK) {
 		mdfour_update(&m, (uchar *)(buf1+i), CSUM_CHUNK);
 	}
@@ -83,7 +83,7 @@ void get_checksum2(char *buf,int len,char *sum)
 	if (len - i > 0 || protocol_version >= 27) {
 		mdfour_update(&m, (uchar *)(buf1+i), (len-i));
 	}
-	
+
 	mdfour_result(&m, (uchar *)sum);
 }
 
@@ -94,35 +94,29 @@ void file_checksum(char *fname,char *sum,OFF_T size)
 	struct map_struct *buf;
 	int fd;
 	OFF_T len = size;
-	char tmpchunk[CSUM_CHUNK];
 	struct mdfour m;
-	
+
 	memset(sum,0,MD4_SUM_LENGTH);
-	
+
 	fd = do_open(fname, O_RDONLY, 0);
-	if (fd == -1) return;
-	
-	buf = map_file(fd,size);
-	
+	if (fd == -1)
+		return;
+
+	buf = map_file(fd, size, MAX_MAP_SIZE, CSUM_CHUNK);
+
 	mdfour_begin(&m);
 
 	for(i = 0; i + CSUM_CHUNK <= len; i += CSUM_CHUNK) {
-		memcpy(tmpchunk, map_ptr(buf,i,CSUM_CHUNK), CSUM_CHUNK);
-		mdfour_update(&m, (uchar *)tmpchunk, CSUM_CHUNK);
+		mdfour_update(&m, (uchar *)map_ptr(buf, i, CSUM_CHUNK),
+			      CSUM_CHUNK);
 	}
 
-	/*
-	 * Prior to version 27 an incorrect MD4 checksum was computed
+	/* Prior to version 27 an incorrect MD4 checksum was computed
 	 * by failing to call mdfour_tail() for block sizes that
 	 * are multiples of 64.  This is fixed by calling mdfour_update()
-	 * even when there are no more bytes.
-	 */
-	if (len - i > 0) {
-		memcpy(tmpchunk, map_ptr(buf,i,len-i), len-i);
-	}
-	if (len - i > 0 || protocol_version >= 27) {
-		mdfour_update(&m, (uchar *)tmpchunk, (len-i));
-	}
+	 * even when there are no more bytes. */
+	if (len - i > 0 || protocol_version >= 27)
+		mdfour_update(&m, (uchar *)map_ptr(buf, i, len-i), len-i);
 
 	mdfour_result(&m, (uchar *)sum);
 
@@ -135,13 +129,13 @@ static int sumresidue;
 static char sumrbuf[CSUM_CHUNK];
 static struct mdfour md;
 
-void sum_init(void)
+void sum_init(int seed)
 {
 	char s[4];
 	mdfour_begin(&md);
-	sumresidue=0;
-	SIVAL(s,0,checksum_seed);
-	sum_update(s,4);
+	sumresidue = 0;
+	SIVAL(s, 0, seed);
+	sum_update(s, 4);
 }
 
 /**
@@ -178,7 +172,7 @@ void sum_update(char *p, int len)
 		sumresidue = len-i;
 		memcpy(sumrbuf,p+i,sumresidue);
 	} else {
-		sumresidue = 0;    
+		sumresidue = 0;
 	}
 }
 
