@@ -1,19 +1,19 @@
 /* -*- c-file-style: "linux" -*-
-   
+
    Copyright (C) 1996-2001 by Andrew Tridgell <tridge@samba.org>
    Copyright (C) Paul Mackerras 1996
    Copyright (C) 2001, 2002 by Martin Pool <mbp@samba.org>
-   
+
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
-   
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-   
+
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
@@ -24,7 +24,45 @@
 time_t starttime = 0;
 
 extern struct stats stats;
+extern int am_root;
+extern int am_server;
+extern int am_sender;
+extern int am_generator;
+extern int am_daemon;
 extern int verbose;
+extern int blocking_io;
+extern int cvs_exclude;
+extern int delete_mode;
+extern int delete_excluded;
+extern int delete_after;
+extern int daemon_over_rsh;
+extern int do_stats;
+extern int dry_run;
+extern int list_only;
+extern int local_server;
+extern int log_got_error;
+extern int module_id;
+extern int orig_umask;
+extern int copy_links;
+extern int keep_dirlinks;
+extern int preserve_hard_links;
+extern int protocol_version;
+extern int recurse;
+extern int relative_paths;
+extern int rsync_port;
+extern int whole_file;
+extern int read_batch;
+extern int write_batch;
+extern int batch_fd;
+extern int batch_gen_fd;
+extern int filesfrom_fd;
+extern pid_t cleanup_child_pid;
+extern char *files_from;
+extern char *remote_filesfrom_file;
+extern char *rsync_path;
+extern char *shell_cmd;
+extern char *batch_name;
+
 
 /* there's probably never more than at most 2 outstanding child processes,
  * but set it higher just in case.
@@ -48,10 +86,10 @@ void wait_process(pid_t pid, int *status)
 
 	while ((waited_pid = waitpid(pid, status, WNOHANG)) == 0) {
 		msleep(20);
-		io_flush();
+		io_flush(FULL_FLUSH);
 	}
-        
-	if ((waited_pid == -1) && (errno == ECHILD)) {
+
+	if (waited_pid == -1 && errno == ECHILD) {
 		/* status of requested child no longer available.
 		 * check to see if it was processed by the sigchld_handler.
 		 */
@@ -64,93 +102,96 @@ void wait_process(pid_t pid, int *status)
 		}
 	}
 
-        /* TODO: If the child exited on a signal, then log an
-         * appropriate error message.  Perhaps we should also accept a
-         * message describing the purpose of the child.  Also indicate
-         * this to the caller so that thhey know something went
-         * wrong.  */
+	/* TODO: If the child exited on a signal, then log an
+	 * appropriate error message.  Perhaps we should also accept a
+	 * message describing the purpose of the child.  Also indicate
+	 * this to the caller so that thhey know something went
+	 * wrong.  */
 	*status = WEXITSTATUS(*status);
 }
 
+/* This function gets called from all 3 processes.  We want the client side
+ * to actually output the text, but the sender is the only process that has
+ * all the stats we need.  So, if we're a client sender, we do the report.
+ * If we're a server sender, we write the stats on the supplied fd.  If
+ * we're the client receiver we read the stats from the supplied fd and do
+ * the report.  All processes might also generate a set of debug stats, if
+ * the verbose level is high enough (this is the only thing that the
+ * generator process and the server receiver ever do here). */
 static void report(int f)
 {
+	/* Cache two stats because the read/write code can change it. */
+	int64 total_read = stats.total_read;
+	int64 total_written = stats.total_written;
 	time_t t = time(NULL);
-	extern int am_server;
-	extern int am_sender;
-	extern int am_daemon;
-	extern int do_stats;
-	extern int remote_version;
-	int send_stats;
 
-	if (do_stats) {
+	if (do_stats && verbose > 1) {
 		/* These come out from every process */
 		show_malloc_stats();
 		show_flist_stats();
 	}
 
+	if (am_generator)
+		return;
+
 	if (am_daemon) {
 		log_exit(0, __FILE__, __LINE__);
-		if (f == -1 || !am_sender) return;
+		if (f == -1 || !am_sender)
+			return;
 	}
 
-	send_stats = verbose || (remote_version >= 20);
 	if (am_server) {
-		if (am_sender && send_stats) {
-			int64 w;
-			/* store total_written in a temporary
-			    because write_longint changes it */
-			w = stats.total_written;
-			write_longint(f,stats.total_read);
-			write_longint(f,w);
-			write_longint(f,stats.total_size);
+		if (am_sender) {
+			write_longint(f, total_read);
+			write_longint(f, total_written);
+			write_longint(f, stats.total_size);
 		}
 		return;
 	}
 
 	/* this is the client */
-	    
-	if (!am_sender && send_stats) {
-		int64 r;
-		stats.total_written = read_longint(f);
-		/* store total_read in a temporary, read_longint changes it */
-		r = read_longint(f);
+
+	if (!am_sender) {
+		/* Read the first two in opposite order because the meaning of
+		 * read/write swaps when switching from sender to receiver. */
+		total_written = read_longint(f);
+		total_read = read_longint(f);
 		stats.total_size = read_longint(f);
-		stats.total_read = r;
+	} else if (write_batch) {
+		/* The --read-batch process is going to be a client
+		 * receiver, so we need to give it the stats. */
+		write_longint(batch_fd, total_read);
+		write_longint(batch_fd, total_written);
+		write_longint(batch_fd, stats.total_size);
 	}
 
 	if (do_stats) {
-		if (!am_sender && !send_stats) {
-		    /* missing the bytes written by the generator */
-		    rprintf(FINFO, "\nCannot show stats as receiver because remote protocol version is less than 20\n");
-		    rprintf(FINFO, "Use --stats -v to show stats\n");
-		    return;
-		}
 		rprintf(FINFO,"\nNumber of files: %d\n", stats.num_files);
-		rprintf(FINFO,"Number of files transferred: %d\n", 
-		       stats.num_transferred_files);
-		rprintf(FINFO,"Total file size: %.0f bytes\n", 
-		       (double)stats.total_size);
-		rprintf(FINFO,"Total transferred file size: %.0f bytes\n", 
-		       (double)stats.total_transferred_size);
-		rprintf(FINFO,"Literal data: %.0f bytes\n", 
-		       (double)stats.literal_data);
-		rprintf(FINFO,"Matched data: %.0f bytes\n", 
-		       (double)stats.matched_data);
+		rprintf(FINFO,"Number of files transferred: %d\n",
+			stats.num_transferred_files);
+		rprintf(FINFO,"Total file size: %.0f bytes\n",
+			(double)stats.total_size);
+		rprintf(FINFO,"Total transferred file size: %.0f bytes\n",
+			(double)stats.total_transferred_size);
+		rprintf(FINFO,"Literal data: %.0f bytes\n",
+			(double)stats.literal_data);
+		rprintf(FINFO,"Matched data: %.0f bytes\n",
+			(double)stats.matched_data);
 		rprintf(FINFO,"File list size: %d\n", stats.flist_size);
-		rprintf(FINFO,"Total bytes written: %.0f\n", 
-		       (double)stats.total_written);
-		rprintf(FINFO,"Total bytes read: %.0f\n\n", 
-		       (double)stats.total_read);
+		rprintf(FINFO,"Total bytes sent: %.0f\n",
+			(double)total_written);
+		rprintf(FINFO,"Total bytes received: %.0f\n",
+			(double)total_read);
 	}
-	
+
 	if (verbose || do_stats) {
-		rprintf(FINFO,"wrote %.0f bytes  read %.0f bytes  %.2f bytes/sec\n",
-		       (double)stats.total_written,
-		       (double)stats.total_read,
-		       (stats.total_written+stats.total_read)/(0.5 + (t-starttime)));
-		rprintf(FINFO,"total size is %.0f  speedup is %.2f\n",
-		       (double)stats.total_size,
-		       (1.0*stats.total_size)/(stats.total_written+stats.total_read));
+		rprintf(FINFO,
+			"\nsent %.0f bytes  received %.0f bytes  %.2f bytes/sec\n",
+			(double)total_written, (double)total_read,
+			(total_written + total_read)/(0.5 + (t - starttime)));
+		rprintf(FINFO, "total size is %.0f  speedup is %.2f\n",
+			(double)stats.total_size,
+			(double)stats.total_size / (total_written+total_read));
 	}
 
 	fflush(stdout);
@@ -165,22 +206,21 @@ static void show_malloc_stats(void)
 {
 #ifdef HAVE_MALLINFO
 	struct mallinfo mi;
-	extern int am_server;
-	extern int am_sender;
-	extern int am_daemon;
 
 	mi = mallinfo();
 
-	rprintf(FINFO, RSYNC_NAME "[%d] (%s%s%s) heap statistics:\n",
+	rprintf(FINFO, "\n" RSYNC_NAME "[%d] (%s%s%s) heap statistics:\n",
 		getpid(),
 		am_server ? "server " : "",
 		am_daemon ? "daemon " : "",
-		am_sender ? "sender" : "receiver");
+		who_am_i());
 	rprintf(FINFO, "  arena:     %10d   (bytes from sbrk)\n", mi.arena);
 	rprintf(FINFO, "  ordblks:   %10d   (chunks not in use)\n", mi.ordblks);
 	rprintf(FINFO, "  smblks:    %10d\n", mi.smblks);
 	rprintf(FINFO, "  hblks:     %10d   (chunks from mmap)\n", mi.hblks);
 	rprintf(FINFO, "  hblkhd:    %10d   (bytes from mmap)\n", mi.hblkhd);
+	rprintf(FINFO, "  allmem:    %10d   (bytes from sbrk + mmap)\n",
+	    mi.arena + mi.hblkhd);
 	rprintf(FINFO, "  usmblks:   %10d\n", mi.usmblks);
 	rprintf(FINFO, "  fsmblks:   %10d\n", mi.fsmblks);
 	rprintf(FINFO, "  uordblks:  %10d   (bytes used)\n", mi.uordblks);
@@ -191,34 +231,36 @@ static void show_malloc_stats(void)
 
 
 /* Start the remote shell.   cmd may be NULL to use the default. */
-static pid_t do_cmd(char *cmd,char *machine,char *user,char *path,int *f_in,int *f_out)
+static pid_t do_cmd(char *cmd, char *machine, char *user, char *path,
+		    int *f_in, int *f_out)
 {
-	char *args[100];
-	int i,argc=0;
+	int i, argc = 0;
+	char *args[MAX_ARGS];
 	pid_t ret;
-	char *tok,*dir=NULL;
+	char *tok, *dir = NULL;
 	int dash_l_set = 0;
-	extern int local_server;
-	extern char *rsync_path;
-	extern int blocking_io;
-	extern int daemon_over_rsh;
-	extern int read_batch;
 
 	if (!read_batch && !local_server) {
+		char *rsh_env = getenv(RSYNC_RSH_ENV);
 		if (!cmd)
-			cmd = getenv(RSYNC_RSH_ENV);
+			cmd = rsh_env;
 		if (!cmd)
 			cmd = RSYNC_RSH;
 		cmd = strdup(cmd);
-		if (!cmd) 
+		if (!cmd)
 			goto oom;
 
-		for (tok=strtok(cmd," ");tok;tok=strtok(NULL," ")) {
+		for (tok = strtok(cmd, " "); tok; tok = strtok(NULL, " ")) {
+			/* Comparison leaves rooms for server_options(). */
+			if (argc >= MAX_ARGS - 100) {
+				rprintf(FERROR, "internal: args[] overflowed in do_cmd()\n");
+				exit_cleanup(RERR_SYNTAX);
+			}
 			args[argc++] = tok;
 		}
 
-		/* check to see if we've already been given '-l user' in 
-		   the remote-shell command */
+		/* check to see if we've already been given '-l user' in
+		 * the remote-shell command */
 		for (i = 0; i < argc-1; i++) {
 			if (!strcmp(args[i], "-l") && args[i+1][0] != '-')
 				dash_l_set = 1;
@@ -241,36 +283,59 @@ static pid_t do_cmd(char *cmd,char *machine,char *user,char *path,int *f_in,int
 
 		args[argc++] = rsync_path;
 
-		if ((blocking_io == -1) && (strcmp(cmd, RSYNC_RSH) == 0))
-			blocking_io = 1;
+		if (blocking_io < 0) {
+			char *cp;
+			if ((cp = strrchr(cmd, '/')) != NULL)
+				cp++;
+			else
+				cp = cmd;
+			if (strcmp(cp, "rsh") == 0 || strcmp(cp, "remsh") == 0)
+				blocking_io = 1;
+		}
 
 		server_options(args,&argc);
 
+		if (argc >= MAX_ARGS - 2) {
+			rprintf(FERROR, "internal: args[] overflowed in do_cmd()\n");
+			exit_cleanup(RERR_SYNTAX);
+		}
 	}
 
 	args[argc++] = ".";
 
-	if (!daemon_over_rsh && path && *path) 
+	if (!daemon_over_rsh && path && *path)
 		args[argc++] = path;
 
 	args[argc] = NULL;
 
 	if (verbose > 3) {
 		rprintf(FINFO,"cmd=");
-		for (i=0;i<argc;i++)
+		for (i = 0; i < argc; i++)
 			rprintf(FINFO,"%s ",args[i]);
 		rprintf(FINFO,"\n");
 	}
 
-	if (local_server) {
-		if (read_batch)
-		    create_flist_from_batch(); /* sets batch_flist */
+	if (read_batch) {
+		int from_gen_pipe[2];
+		if (fd_pair(from_gen_pipe) < 0) {
+			rsyserr(FERROR, errno, "pipe");
+			exit_cleanup(RERR_IPC);
+		}
+		batch_gen_fd = from_gen_pipe[0];
+		*f_out = from_gen_pipe[1];
+		*f_in = batch_fd;
+		ret = -1; /* no child pid */
+	} else if (local_server) {
+		/* If the user didn't request --[no-]whole-file, force
+		 * it on, but only if we're not batch processing. */
+		if (whole_file < 0 && !write_batch)
+			whole_file = 1;
 		ret = local_child(argc, args, f_in, f_out, child_main);
-	} else {
+	} else
 		ret = piped_child(args,f_in,f_out);
-	}
 
-	if (dir) free(dir);
+	if (dir)
+		free(dir);
 
 	return ret;
 
@@ -280,25 +345,23 @@ oom:
 }
 
 
-
-
 static char *get_local_name(struct file_list *flist,char *name)
 {
 	STRUCT_STAT st;
-	extern int orig_umask;
+	int e;
 
 	if (verbose > 2)
-		rprintf(FINFO,"get_local_name count=%d %s\n", 
+		rprintf(FINFO,"get_local_name count=%d %s\n",
 			flist->count, NS(name));
 
-	if (!name) 
+	if (!name)
 		return NULL;
 
 	if (do_stat(name,&st) == 0) {
 		if (S_ISDIR(st.st_mode)) {
-			if (!push_dir(name, 0)) {
-				rprintf(FERROR,"push_dir %s : %s (1)\n",
-					name,strerror(errno));
+			if (!push_dir(name)) {
+				rsyserr(FERROR, errno, "push_dir#1 %s failed",
+					full_fname(name));
 				exit_cleanup(RERR_FILESELECT);
 			}
 			return NULL;
@@ -310,21 +373,24 @@ static char *get_local_name(struct file_list *flist,char *name)
 		return name;
 	}
 
-	if (flist->count <= 1)
+	if (flist->count <= 1 && ((e = strlen(name)) <= 1 || name[e-1] != '/'))
 		return name;
 
 	if (do_mkdir(name,0777 & ~orig_umask) != 0) {
-		rprintf(FERROR, RSYNC_NAME ": mkdir %s: %s\n",
-			name, strerror(errno));
+		rsyserr(FERROR, errno, "mkdir %s failed", full_fname(name));
 		exit_cleanup(RERR_FILEIO);
-	} else {
-		if (verbose > 0)
-			rprintf(FINFO,"created directory %s\n",name);
+	}
+	if (verbose > 0)
+		rprintf(FINFO, "created directory %s\n", name);
+
+	if (dry_run) {
+		dry_run++;
+		return NULL;
 	}
 
-	if (!push_dir(name, 0)) {
-		rprintf(FERROR, RSYNC_NAME ": push_dir %s: %s\n",
-			name, strerror(errno));
+	if (!push_dir(name)) {
+		rsyserr(FERROR, errno, "push_dir#2 %s failed",
+			full_fname(name));
 		exit_cleanup(RERR_FILESELECT);
 	}
 
@@ -332,54 +398,61 @@ static char *get_local_name(struct file_list *flist,char *name)
 }
 
 
-
-
 static void do_server_sender(int f_in, int f_out, int argc,char *argv[])
 {
 	int i;
 	struct file_list *flist;
 	char *dir = argv[0];
-	extern int relative_paths;
-	extern int recurse;
-	extern int remote_version;
 
-	if (verbose > 2)
-		rprintf(FINFO,"server_sender starting pid=%d\n",(int)getpid());
-  
-	if (!relative_paths && !push_dir(dir, 0)) {
-		rprintf(FERROR,"push_dir %s: %s (3)\n",dir,strerror(errno));
+	if (verbose > 2) {
+		rprintf(FINFO, "server_sender starting pid=%ld\n",
+			(long)getpid());
+	}
+
+	if (am_daemon && lp_write_only(module_id)) {
+		rprintf(FERROR, "ERROR: module is write only\n");
+		exit_cleanup(RERR_SYNTAX);
+		return;
+	}
+
+	if (!relative_paths && !push_dir(dir)) {
+		rsyserr(FERROR, errno, "push_dir#3 %s failed",
+			full_fname(dir));
 		exit_cleanup(RERR_FILESELECT);
 	}
 	argc--;
 	argv++;
-  
+
 	if (strcmp(dir,".")) {
 		int l = strlen(dir);
-		if (strcmp(dir,"/") == 0) 
+		if (strcmp(dir,"/") == 0)
 			l = 0;
-		for (i=0;i<argc;i++)
+		for (i = 0; i < argc; i++)
 			argv[i] += l+1;
 	}
 
 	if (argc == 0 && recurse) {
-		argc=1;
+		argc = 1;
 		argv--;
 		argv[0] = ".";
 	}
-	
+
 	flist = send_file_list(f_out,argc,argv);
 	if (!flist || flist->count == 0) {
 		exit_cleanup(0);
 	}
 
+	io_start_buffering_in();
+	io_start_buffering_out();
+
 	send_files(flist,f_out,f_in);
-	io_flush();
+	io_flush(FULL_FLUSH);
 	report(f_out);
-	if (remote_version >= 24) {
-		/* final goodbye message */		
- 		read_int(f_in);
- 	}
-	io_flush();
+	if (protocol_version >= 24) {
+		/* final goodbye message */
+		read_int(f_in);
+	}
+	io_flush(FULL_FLUSH);
 	exit_cleanup(0);
 }
 
@@ -387,81 +460,78 @@ static void do_server_sender(int f_in, int f_out, int argc,char *argv[])
 static int do_recv(int f_in,int f_out,struct file_list *flist,char *local_name)
 {
 	int pid;
-	int status=0;
-	int recv_pipe[2];
+	int status = 0;
 	int error_pipe[2];
-	extern int preserve_hard_links;
-	extern int delete_after;
-	extern int recurse;
-	extern int delete_mode;
-	extern int remote_version;
+
+	/* The receiving side mustn't obey this, or an existing symlink that
+	 * points to an identical file won't be replaced by the referent. */
+	copy_links = 0;
 
 	if (preserve_hard_links)
 		init_hard_links(flist);
 
 	if (!delete_after) {
 		/* I moved this here from recv_files() to prevent a race condition */
-		if (recurse && delete_mode && !local_name && flist->count>0) {
+		if (recurse && delete_mode && !local_name && flist->count > 0)
 			delete_files(flist);
-		}
-	}
-
-	if (fd_pair(recv_pipe) < 0) {
-		rprintf(FERROR,"pipe failed in do_recv\n");
-		exit_cleanup(RERR_SOCKETIO);
 	}
 
 	if (fd_pair(error_pipe) < 0) {
 		rprintf(FERROR,"error pipe failed in do_recv\n");
 		exit_cleanup(RERR_SOCKETIO);
 	}
-  
-	io_flush();
 
-	if ((pid=do_fork()) == 0) {
-		close(recv_pipe[0]);
+	io_flush(NORMAL_FLUSH);
+
+	if ((pid = do_fork()) == 0) {
 		close(error_pipe[0]);
-		if (f_in != f_out) close(f_out);
+		if (f_in != f_out)
+			close(f_out);
 
 		/* we can't let two processes write to the socket at one time */
-		io_multiplexing_close();
+		close_multiplexing_out();
 
 		/* set place to send errors */
-		set_error_fd(error_pipe[1]);
+		set_msg_fd_out(error_pipe[1]);
 
-		recv_files(f_in,flist,local_name,recv_pipe[1]);
-		io_flush();
+		recv_files(f_in,flist,local_name);
+		io_flush(FULL_FLUSH);
 		report(f_in);
 
-		write_int(recv_pipe[1],1);
-		close(recv_pipe[1]);
-		io_flush();
+		send_msg(MSG_DONE, "", 0);
+		io_flush(FULL_FLUSH);
 		/* finally we go to sleep until our parent kills us
-		   with a USR2 signal. We sleep for a short time as on
-		   some OSes a signal won't interrupt a sleep! */
-		while (msleep(20))
-			;
+		 * with a USR2 signal. We sleep for a short time as on
+		 * some OSes a signal won't interrupt a sleep! */
+		while (1)
+			msleep(20);
 	}
 
-	close(recv_pipe[1]);
+	am_generator = 1;
+	close_multiplexing_in();
+	if (write_batch)
+		stop_write_batch();
+
 	close(error_pipe[1]);
-	if (f_in != f_out) close(f_in);
+	if (f_in != f_out)
+		close(f_in);
 
-	io_start_buffering(f_out);
+	io_start_buffering_out();
 
-	io_set_error_fd(error_pipe[0]);
+	set_msg_fd_in(error_pipe[0]);
 
-	generate_files(f_out,flist,local_name,recv_pipe[0]);
+	generate_files(f_out, flist, local_name);
 
-	read_int(recv_pipe[0]);
-	close(recv_pipe[0]);
-	if (remote_version >= 24) {
+	get_redo_num(); /* Read final MSG_DONE and any prior messages. */
+	report(-1);
+	io_flush(FULL_FLUSH);
+	if (protocol_version >= 24) {
 		/* send a final goodbye message */
 		write_int(f_out, -1);
 	}
-	io_flush();
+	io_flush(FULL_FLUSH);
 
-	io_set_error_fd(-1);
+	set_msg_fd_in(-1);
 	kill(pid, SIGUSR2);
 	wait_process(pid, &status);
 	return status;
@@ -472,53 +542,56 @@ static void do_server_recv(int f_in, int f_out, int argc,char *argv[])
 {
 	int status;
 	struct file_list *flist;
-	char *local_name=NULL;
+	char *local_name = NULL;
 	char *dir = NULL;
-	extern int delete_mode;
-	extern int delete_excluded;
-	extern int am_daemon;
-	extern int module_id;
-	extern int am_sender;
-	extern int read_batch;
-	extern struct file_list *batch_flist;
 
-	if (verbose > 2)
-		rprintf(FINFO,"server_recv(%d) starting pid=%d\n",argc,(int)getpid());
+	if (verbose > 2) {
+		rprintf(FINFO, "server_recv(%d) starting pid=%ld\n",
+			argc, (long)getpid());
+	}
 
-	if (am_daemon && lp_read_only(module_id) && !am_sender) {
+	if (am_daemon && lp_read_only(module_id)) {
 		rprintf(FERROR,"ERROR: module is read only\n");
 		exit_cleanup(RERR_SYNTAX);
 		return;
 	}
 
-	
+
 	if (argc > 0) {
 		dir = argv[0];
 		argc--;
 		argv++;
-		if (!am_daemon && !push_dir(dir, 0)) {
-			rprintf(FERROR,"push_dir %s : %s (4)\n",
-				dir,strerror(errno));
+		if (!am_daemon && !push_dir(dir)) {
+			rsyserr(FERROR, errno, "push_dir#4 %s failed",
+				full_fname(dir));
 			exit_cleanup(RERR_FILESELECT);
-		}    
+		}
 	}
 
+	io_start_buffering_in();
 	if (delete_mode && !delete_excluded)
 		recv_exclude_list(f_in);
 
-	if (read_batch)
-	    flist = batch_flist;
-	else
-	    flist = recv_file_list(f_in);
+	if (filesfrom_fd >= 0) {
+		/* We're receiving the file info from the sender, so we need
+		 * the IO routines to automatically write out the names onto
+		 * our f_out socket as we read the list info from the sender.
+		 * This avoids both deadlock and extra delays/buffers. */
+		io_set_filesfrom_fds(filesfrom_fd, f_out);
+		filesfrom_fd = -1;
+	}
+
+	flist = recv_file_list(f_in);
 	if (!flist) {
 		rprintf(FERROR,"server_recv: recv_file_list error\n");
 		exit_cleanup(RERR_FILESELECT);
 	}
-	
-	if (argc > 0) {    
+
+	if (argc > 0) {
 		if (strcmp(dir,".")) {
 			argv[0] += strlen(dir);
-			if (argv[0][0] == '/') argv[0]++;
+			if (argv[0][0] == '/')
+				argv[0]++;
 		}
 		local_name = get_local_name(flist,argv[0]);
 	}
@@ -537,25 +610,21 @@ int child_main(int argc, char *argv[])
 
 void start_server(int f_in, int f_out, int argc, char *argv[])
 {
-	extern int cvs_exclude;
-	extern int am_sender;
-	extern int remote_version;
-	extern int read_batch;
-
-	setup_protocol(f_out, f_in);
-
 	set_nonblocking(f_in);
 	set_nonblocking(f_out);
 
-	if (remote_version >= 23)
-		io_start_multiplex_out(f_out);
+	io_set_sock_fds(f_in, f_out);
+	setup_protocol(f_out, f_in);
+
+	if (protocol_version >= 23)
+		io_start_multiplex_out();
 
 	if (am_sender) {
-		if (!read_batch) {
-		    recv_exclude_list(f_in);
-		    if (cvs_exclude)
+		keep_dirlinks = 0; /* Must be disabled on the sender. */
+
+		recv_exclude_list(f_in);
+		if (cvs_exclude)
 			add_cvs_excludes();
-		}
 		do_server_sender(f_in, f_out, argc, argv);
 	} else {
 		do_server_recv(f_in, f_out, argc, argv);
@@ -573,99 +642,101 @@ int client_run(int f_in, int f_out, pid_t pid, int argc, char *argv[])
 	struct file_list *flist = NULL;
 	int status = 0, status2 = 0;
 	char *local_name = NULL;
-	extern int am_sender;
-	extern int remote_version;
-	extern pid_t cleanup_child_pid;
-	extern int write_batch;
-	extern int read_batch;
-	extern struct file_list *batch_flist;
 
 	cleanup_child_pid = pid;
-	if (read_batch)
-	    flist = batch_flist;
-
-	set_nonblocking(f_in);
-	set_nonblocking(f_out);
+	if (read_batch) {
+		assert(am_sender == 0);
+	} else {
+		set_nonblocking(f_in);
+		set_nonblocking(f_out);
+	}
 
+	io_set_sock_fds(f_in, f_out);
 	setup_protocol(f_out,f_in);
 
-	if (remote_version >= 23)
-		io_start_multiplex_in(f_in);
-	
+	if (protocol_version >= 23 && !read_batch)
+		io_start_multiplex_in();
+
+	/* We set our stderr file handle to blocking because ssh might have
+	 * set it to non-blocking.  This can be particularly troublesome if
+	 * stderr is a clone of stdout, because ssh would have set our stdout
+	 * to non-blocking at the same time (which can easily cause us to lose
+	 * output from our print statements).  This kluge shouldn't cause ssh
+	 * any problems for how we use it.  Note also that we delayed setting
+	 * this until after the above protocol setup so that we know for sure
+	 * that ssh is done twiddling its file descriptors.  */
+	set_blocking(STDERR_FILENO);
+
 	if (am_sender) {
-		extern int cvs_exclude;
-		extern int delete_mode;
-		extern int delete_excluded;
+		keep_dirlinks = 0; /* Must be disabled on the sender. */
+		io_start_buffering_out();
 		if (cvs_exclude)
 			add_cvs_excludes();
-		if (delete_mode && !delete_excluded) 
+		if (delete_mode && !delete_excluded)
 			send_exclude_list(f_out);
-		if (!read_batch) /*  dw -- don't write to pipe */
-		    flist = send_file_list(f_out,argc,argv);
-		if (verbose > 3) 
+		if (remote_filesfrom_file)
+			filesfrom_fd = f_in;
+
+		if (write_batch)
+			start_write_batch(f_out);
+		if (!read_batch) /* don't write to pipe */
+			flist = send_file_list(f_out,argc,argv);
+		if (verbose > 3)
 			rprintf(FINFO,"file list sent\n");
 
+		io_flush(NORMAL_FLUSH);
 		send_files(flist,f_out,f_in);
-		if (remote_version >= 24) {
-			/* final goodbye message */		
+		io_flush(FULL_FLUSH);
+		if (protocol_version >= 24) {
+			/* final goodbye message */
 			read_int(f_in);
 		}
 		if (pid != -1) {
 			if (verbose > 3)
 				rprintf(FINFO,"client_run waiting on %d\n", (int) pid);
-			io_flush();
+			io_flush(FULL_FLUSH);
 			wait_process(pid, &status);
 		}
 		report(-1);
+		io_flush(FULL_FLUSH);
 		exit_cleanup(status);
 	}
 
-	if (argc == 0) {
-		extern int list_only;
+	if (argc == 0)
 		list_only = 1;
+
+	if (!read_batch)
+		send_exclude_list(f_out);
+
+	if (filesfrom_fd >= 0) {
+		io_set_filesfrom_fds(filesfrom_fd, f_out);
+		filesfrom_fd = -1;
 	}
-	
-	if (!write_batch)
-	    send_exclude_list(f_out);
-	
+
+	if (write_batch)
+		start_write_batch(f_in);
 	flist = recv_file_list(f_in);
 	if (!flist || flist->count == 0) {
 		rprintf(FINFO, "client: nothing to do: "
-                        "perhaps you need to specify some filenames or "
-                        "the --recursive option?\n");
+			"perhaps you need to specify some filenames or "
+			"the --recursive option?\n");
 		exit_cleanup(0);
 	}
-	
+
 	local_name = get_local_name(flist,argv[0]);
-	
+
 	status2 = do_recv(f_in,f_out,flist,local_name);
-	
+
 	if (pid != -1) {
 		if (verbose > 3)
 			rprintf(FINFO,"client_run2 waiting on %d\n", (int) pid);
-		io_flush();
+		io_flush(FULL_FLUSH);
 		wait_process(pid, &status);
 	}
-	
+
 	return MAX(status, status2);
 }
 
-static char *find_colon(char *s)
-{
-	char *p, *p2;
-
-	p = strchr(s,':');
-	if (!p) return NULL;
-	
-	/* now check to see if there is a / in the string before the : - if there is then
-	   discard the colon on the assumption that the : is part of a filename */
-	p2 = strchr(s,'/');
-	if (p2 && p2 < p) return NULL;
-
-	return p;
-}
-
-
 static int copy_argv (char *argv[])
 {
 	int i;
@@ -699,123 +770,148 @@ static int start_client(int argc, char *argv[])
 	int ret;
 	pid_t pid;
 	int f_in,f_out;
-	extern int local_server;
-	extern int am_sender;
-	extern char *shell_cmd;
-	extern int rsync_port;
-	extern int daemon_over_rsh;
-	extern int read_batch;
 	int rc;
 
 	/* Don't clobber argv[] so that ps(1) can still show the right
-           command line. */
+	 * command line. */
 	if ((rc = copy_argv(argv)))
 		return rc;
 
 	/* rsync:// always uses rsync server over direct socket connection */
-	if (strncasecmp(URL_PREFIX, argv[0], strlen(URL_PREFIX)) == 0) {
+	if (strncasecmp(URL_PREFIX, argv[0], strlen(URL_PREFIX)) == 0
+	    && !read_batch) {
 		char *host, *path;
 
 		host = argv[0] + strlen(URL_PREFIX);
 		p = strchr(host,'/');
 		if (p) {
-			*p = 0;
+			*p = '\0';
 			path = p+1;
-		} else {
+		} else
 			path = "";
-		}
-		p = strchr(host,':');
+		if (*host == '[' && (p = strchr(host, ']')) != NULL) {
+			host++;
+			*p++ = '\0';
+			if (*p != ':')
+				p = NULL;
+		} else
+			p = strchr(host, ':');
 		if (p) {
 			rsync_port = atoi(p+1);
-			*p = 0;
+			*p = '\0';
 		}
 		return start_socket_client(host, path, argc-1, argv+1);
 	}
 
-	if (!read_batch) {
+	if (!read_batch) { /* for read_batch, NO source is specified */
 		p = find_colon(argv[0]);
-
-	if (p) {
-		if (p[1] == ':') { /* double colon */
-			*p = 0;
-			if (!shell_cmd) {
-				return start_socket_client(argv[0], p+2,
-							   argc-1, argv+1);
+		if (p) { /* source is remote */
+			if (remote_filesfrom_file
+			 && remote_filesfrom_file != files_from + 1
+			 && strncmp(files_from, argv[0], p-argv[0]+1) != 0) {
+				rprintf(FERROR,
+					"--files-from hostname is not the same as the transfer hostname\n");
+				exit_cleanup(RERR_SYNTAX);
 			}
-			p++;
-			daemon_over_rsh = 1;
-		}
-
-		if (argc < 1) {
-			usage(FERROR);
-			exit_cleanup(RERR_SYNTAX);
-		}
-
-		am_sender = 0;
-		*p = 0;
-		shell_machine = argv[0];
-		shell_path = p+1;
-		argc--;
-		argv++;
-	} else {
-		am_sender = 1;
-
-		/* rsync:// destination uses rsync server over direct socket */
-		if (strncasecmp(URL_PREFIX, argv[argc-1], strlen(URL_PREFIX)) == 0) {
-			char *host, *path;
-
-			host = argv[argc-1] + strlen(URL_PREFIX);
-			p = strchr(host,'/');
-			if (p) {
+			if (p[1] == ':') { /* double colon */
 				*p = 0;
-				path = p+1;
-			} else {
-				path = "";
+				if (!shell_cmd) {
+					return start_socket_client(argv[0], p+2,
+								   argc-1, argv+1);
+				}
+				p++;
+				daemon_over_rsh = 1;
 			}
-			p = strchr(host,':');
-			if (p) {
-				rsync_port = atoi(p+1);
-				*p = 0;
-			}
-			return start_socket_client(host, path, argc-1, argv);
-		}
 
-		p = find_colon(argv[argc-1]);
-		if (!p) {
-			local_server = 1;
-		} else if (p[1] == ':') { /* double colon */
-			*p = 0;
-			if (!shell_cmd) {
-				return start_socket_client(argv[argc-1], p+2,
-							   argc-1, argv);
+			if (argc < 1) { /* destination required */
+				usage(FERROR);
+				exit_cleanup(RERR_SYNTAX);
 			}
-			p++;
-			daemon_over_rsh = 1;
-		}
 
-		if (argc < 2) {
-			usage(FERROR);
-			exit_cleanup(RERR_SYNTAX);
-		}
-		
-		if (local_server) {
-			shell_machine = NULL;
-			shell_path = argv[argc-1];
-		} else {
+			am_sender = 0;
 			*p = 0;
-			shell_machine = argv[argc-1];
+			shell_machine = argv[0];
 			shell_path = p+1;
+			argv++;
+		} else { /* source is local */
+			am_sender = 1;
+
+			/* rsync:// destination uses rsync server over direct socket */
+			if (strncasecmp(URL_PREFIX, argv[argc-1], strlen(URL_PREFIX)) == 0) {
+				char *host, *path;
+
+				host = argv[argc-1] + strlen(URL_PREFIX);
+				p = strchr(host,'/');
+				if (p) {
+					*p = '\0';
+					path = p+1;
+				} else
+					path = "";
+				if (*host == '[' && (p = strchr(host, ']')) != NULL) {
+					host++;
+					*p++ = '\0';
+					if (*p != ':')
+						p = NULL;
+				} else
+					p = strchr(host, ':');
+				if (p) {
+					rsync_port = atoi(p+1);
+					*p = '\0';
+				}
+				return start_socket_client(host, path, argc-1, argv);
+			}
+
+			p = find_colon(argv[argc-1]); /* look in dest arg */
+			if (p && remote_filesfrom_file
+			 && remote_filesfrom_file != files_from + 1
+			 && strncmp(files_from, argv[argc-1], p-argv[argc-1]+1) != 0) {
+				rprintf(FERROR,
+					"--files-from hostname is not the same as the transfer hostname\n");
+				exit_cleanup(RERR_SYNTAX);
+			}
+			if (!p) { /* no colon found, so src & dest are local */
+				local_server = 1;
+				if (remote_filesfrom_file) {
+					rprintf(FERROR,
+						"--files-from cannot be remote when the transfer is local\n");
+					exit_cleanup(RERR_SYNTAX);
+				}
+			} else if (p[1] == ':') { /* double colon */
+				*p = 0;
+				if (!shell_cmd) {
+					return start_socket_client(argv[argc-1], p+2,
+								   argc-1, argv);
+				}
+				p++;
+				daemon_over_rsh = 1;
+			}
+
+			if (argc < 2) {
+				usage(FERROR);
+				exit_cleanup(RERR_SYNTAX);
+			}
+
+			if (local_server) {
+				shell_machine = NULL;
+				shell_path = argv[argc-1];
+			} else {
+				*p = 0;
+				shell_machine = argv[argc-1];
+				shell_path = p+1;
+			}
 		}
 		argc--;
-	}
-	} else {
-	    am_sender = 1;
-	    local_server = 1;
-	    shell_path = argv[argc-1];
+	} else {  /* read_batch */
+		local_server = 1;
+		shell_path = argv[argc-1];
+		if (find_colon(shell_path)) {
+			rprintf(FERROR, "remote destination is not allowed with --read-batch\n");
+			exit_cleanup(RERR_SYNTAX);
+		}
 	}
 
 	if (shell_machine) {
-		p = strchr(shell_machine,'@');
+		p = strrchr(shell_machine,'@');
 		if (p) {
 			*p = 0;
 			shell_user = shell_machine;
@@ -830,22 +926,23 @@ static int start_client(int argc, char *argv[])
 			shell_user?shell_user:"",
 			shell_path?shell_path:"");
 	}
-	
+
+	/* for remote source, only single dest arg can remain ... */
 	if (!am_sender && argc > 1) {
 		usage(FERROR);
 		exit_cleanup(RERR_SYNTAX);
 	}
 
-	if (argc == 0 && !am_sender) {
-		extern int list_only;
+	/* ... or no dest at all */
+	if (!am_sender && argc == 0) {
 		list_only = 1;
 	}
-	
+
 	pid = do_cmd(shell_cmd,shell_machine,shell_user,shell_path,
 		     &f_in,&f_out);
 
 	/* if we're running an rsync server on the remote host over a
-	   remote shell command, we need to do the RSYNCD protocol first */
+	 * remote shell command, we need to do the RSYNCD protocol first */
 	if (daemon_over_rsh) {
 		int tmpret;
 		tmpret = start_inband_exchange(shell_user, shell_path,
@@ -863,35 +960,37 @@ static int start_client(int argc, char *argv[])
 }
 
 
-static RETSIGTYPE sigusr1_handler(int UNUSED(val)) {
+static RETSIGTYPE sigusr1_handler(UNUSED(int val))
+{
 	exit_cleanup(RERR_SIGNAL);
 }
 
-static RETSIGTYPE sigusr2_handler(int UNUSED(val)) {
-	extern int log_got_error;
+static RETSIGTYPE sigusr2_handler(UNUSED(int val))
+{
 	if (log_got_error) _exit(RERR_PARTIAL);
 	_exit(0);
 }
 
-static RETSIGTYPE sigchld_handler(int UNUSED(val)) {
+static RETSIGTYPE sigchld_handler(UNUSED(int val))
+{
 #ifdef WNOHANG
 	int cnt, status;
 	pid_t pid;
 	/* An empty waitpid() loop was put here by Tridge and we could never
-	 * get him to explain why he put it in, so rather than taking it 
+	 * get him to explain why he put it in, so rather than taking it
 	 * out we're instead saving the child exit statuses for later use.
 	 * The waitpid() loop presumably eliminates all possibility of leaving
 	 * zombie children, maybe that's why he did it.
 	 */
 	while ((pid = waitpid(-1, &status, WNOHANG)) > 0) {
-		 /* save the child's exit status */
-		 for (cnt = 0; cnt < MAXCHILDPROCS; cnt++) {
-			  if (pid_stat_table[cnt].pid == 0) {
-				   pid_stat_table[cnt].pid = pid;
-				   pid_stat_table[cnt].status = status;
-				   break;
-			  }
-		 }
+		/* save the child's exit status */
+		for (cnt = 0; cnt < MAXCHILDPROCS; cnt++) {
+			if (pid_stat_table[cnt].pid == 0) {
+				pid_stat_table[cnt].pid = pid;
+				pid_stat_table[cnt].status = status;
+				break;
+			}
+		}
 	}
 #endif
 }
@@ -931,7 +1030,7 @@ const char *get_panic_action(void)
  * should just look at the environment variable, but I'm a bit leery
  * of a signal sending us into a busy loop.
  **/
-static RETSIGTYPE rsync_panic_handler(int UNUSED(whatsig))
+static RETSIGTYPE rsync_panic_handler(UNUSED(int whatsig))
 {
 	char cmd_buf[300];
 	int ret;
@@ -949,19 +1048,10 @@ static RETSIGTYPE rsync_panic_handler(int UNUSED(whatsig))
 
 
 int main(int argc,char *argv[])
-{       
-	extern int am_root;
-	extern int orig_umask;
-	extern int dry_run;
-	extern int am_daemon;
-	extern int am_server;
+{
 	int ret;
-	extern int write_batch;
-	int orig_argc;
-	char **orig_argv;
-
-	orig_argc = argc;
-	orig_argv = argv;
+	int orig_argc = argc;
+	char **orig_argv = argv;
 
 	signal(SIGUSR1, sigusr1_handler);
 	signal(SIGUSR2, sigusr2_handler);
@@ -974,7 +1064,7 @@ int main(int argc,char *argv[])
 #endif /* def MAINTAINER_MODE */
 
 	starttime = time(NULL);
-	am_root = (getuid() == 0);
+	am_root = (MY_UID() == 0);
 
 	memset(&stats, 0, sizeof(stats));
 
@@ -984,12 +1074,12 @@ int main(int argc,char *argv[])
 	}
 
 	/* we set a 0 umask so that correct file permissions can be
-	   carried across */
+	 * carried across */
 	orig_umask = (int)umask(0);
 
 	if (!parse_arguments(&argc, (const char ***) &argv, 1)) {
-                /* FIXME: We ought to call the same error-handling
-                 * code here, rather than relying on getopt. */
+		/* FIXME: We ought to call the same error-handling
+		 * code here, rather than relying on getopt. */
 		option_error();
 		exit_cleanup(RERR_SYNTAX);
 	}
@@ -1003,13 +1093,31 @@ int main(int argc,char *argv[])
 	signal(SIGPIPE, SIG_IGN);
 
 	/* Initialize push_dir here because on some old systems getcwd
-	   (implemented by forking "pwd" and reading its output) doesn't
-	   work when there are other child processes.  Also, on all systems
-	   that implement getcwd that way "pwd" can't be found after chroot. */
-	push_dir(NULL,0);
+	 * (implemented by forking "pwd" and reading its output) doesn't
+	 * work when there are other child processes.  Also, on all systems
+	 * that implement getcwd that way "pwd" can't be found after chroot. */
+	push_dir(NULL);
 
-	if (write_batch && !am_server) {
-	    write_batch_argvs_file(orig_argc, orig_argv);
+	init_flist();
+
+	if (write_batch || read_batch) {
+		if (write_batch)
+			write_batch_shell_file(orig_argc, orig_argv, argc);
+
+		if (read_batch && strcmp(batch_name, "-") == 0)
+			batch_fd = STDIN_FILENO;
+		else {
+			batch_fd = do_open(batch_name,
+				   write_batch ? O_WRONLY | O_CREAT | O_TRUNC
+				   : O_RDONLY, S_IRUSR | S_IWUSR);
+		}
+		if (batch_fd < 0) {
+			rsyserr(FERROR, errno, "Batch file %s open error",
+				batch_name);
+			exit_cleanup(RERR_FILEIO);
+		}
+		if (read_batch)
+			read_stream_flags(batch_fd);
 	}
 
 	if (am_daemon && !am_server)
@@ -1023,13 +1131,6 @@ int main(int argc,char *argv[])
 	if (dry_run)
 		verbose = MAX(verbose,1);
 
-#ifndef SUPPORT_LINKS
-	if (!am_server && preserve_links) {
-		rprintf(FERROR,"ERROR: symbolic links not supported\n");
-		exit_cleanup(RERR_UNSUPPORTED);
-	}
-#endif
-
 	if (am_server) {
 		set_nonblocking(STDIN_FILENO);
 		set_nonblocking(STDOUT_FILENO);
@@ -1039,11 +1140,10 @@ int main(int argc,char *argv[])
 	}
 
 	ret = start_client(argc, argv);
-	if (ret == -1) 
+	if (ret == -1)
 		exit_cleanup(RERR_STARTCLIENT);
 	else
 		exit_cleanup(ret);
 
-	exit(ret);
-	/* NOTREACHED */
+	return ret;
 }
