@@ -21,7 +21,10 @@
 #include "rsync.h"
 #include "popt.h"
 
-extern struct exclude_struct **exclude_list;
+extern int sanitize_paths;
+extern int select_timeout;
+extern struct exclude_list_struct exclude_list;
+extern struct exclude_list_struct server_exclude_list;
 
 int make_backups = 0;
 
@@ -36,6 +39,7 @@ int make_backups = 0;
 int whole_file = -1;
 
 int archive_mode = 0;
+int keep_dirlinks = 0;
 int copy_links = 0;
 int preserve_links = 0;
 int preserve_hard_links = 0;
@@ -46,17 +50,17 @@ int preserve_gid = 0;
 int preserve_times = 0;
 int update_only = 0;
 int cvs_exclude = 0;
-int dry_run=0;
-int local_server=0;
-int ignore_times=0;
-int delete_mode=0;
-int delete_excluded=0;
-int one_file_system=0;
+int dry_run = 0;
+int local_server = 0;
+int ignore_times = 0;
+int delete_mode = 0;
+int delete_excluded = 0;
+int one_file_system = 0;
 int protocol_version = PROTOCOL_VERSION;
-int sparse_files=0;
-int do_compression=0;
-int am_root=0;
-int orig_umask=0;
+int sparse_files = 0;
+int do_compression = 0;
+int am_root = 0;
+int orig_umask = 0;
 int relative_paths = -1;
 int implied_dirs = 1;
 int numeric_ids = 0;
@@ -66,6 +70,7 @@ int read_only = 0;
 int module_id = -1;
 int am_server = 0;
 int am_sender = 0;
+int am_generator = 0;
 char *files_from = NULL;
 int filesfrom_fd = -1;
 char *remote_filesfrom_file = NULL;
@@ -73,21 +78,24 @@ int eol_nulls = 0;
 int recurse = 0;
 int am_daemon = 0;
 int daemon_over_rsh = 0;
-int do_stats=0;
-int do_progress=0;
-int keep_partial=0;
-int safe_symlinks=0;
-int copy_unsafe_links=0;
-int block_size=0;
-int size_only=0;
-int bwlimit=0;
-int delete_after=0;
-int only_existing=0;
-int opt_ignore_existing=0;
-int max_delete=0;
-int ignore_errors=0;
-int modify_window=0;
-int blocking_io=-1;
+int do_stats = 0;
+int do_progress = 0;
+int keep_partial = 0;
+int safe_symlinks = 0;
+int copy_unsafe_links = 0;
+int size_only = 0;
+int bwlimit = 0;
+size_t bwlimit_writemax = 0;
+int delete_after = 0;
+int only_existing = 0;
+int opt_ignore_existing = 0;
+int max_delete = 0;
+int ignore_errors = 0;
+int modify_window = 0;
+int blocking_io = -1;
+int checksum_seed = 0;
+int inplace = 0;
+unsigned int block_size = 0;
 
 
 /** Network address family. **/
@@ -106,9 +114,11 @@ int write_batch = 0;
 int read_batch = 0;
 int backup_dir_len = 0;
 int backup_suffix_len;
+unsigned int backup_dir_remainder;
 
 char *backup_suffix = NULL;
 char *tmpdir = NULL;
+char *partial_dir = NULL;
 char *compare_dest = NULL;
 char *config_file = NULL;
 char *shell_cmd = NULL;
@@ -116,6 +126,7 @@ char *log_format = NULL;
 char *password_file = NULL;
 char *rsync_path = RSYNC_PATH;
 char *backup_dir = NULL;
+char backup_dir_buf[MAXPATHLEN];
 int rsync_port = RSYNC_PORT;
 int link_dest = 0;
 
@@ -124,12 +135,14 @@ int quiet = 0;
 int always_checksum = 0;
 int list_only = 0;
 
-char *batch_prefix = NULL;
+#define MAX_BATCH_NAME_LEN 256	/* Must be less than MAXPATHLEN-13 */
+char *batch_name = NULL;
 
+static int daemon_opt;   /* sets am_daemon after option error-reporting */
 static int modify_window_set;
 
 /** Local address to bind.  As a character string because it's
- * interpreted by the IPv6 layer: should be a numeric IP4 or ip6
+ * interpreted by the IPv6 layer: should be a numeric IP4 or IP6
  * address, or a hostname. **/
 char *bind_address;
 
@@ -137,6 +150,7 @@ char *bind_address;
 static void print_rsync_version(enum logcode f)
 {
 	char const *got_socketpair = "no ";
+	char const *have_inplace = "no ";
 	char const *hardlinks = "no ";
 	char const *links = "no ";
 	char const *ipv6 = "no ";
@@ -146,6 +160,10 @@ static void print_rsync_version(enum logcode f)
 	got_socketpair = "";
 #endif
 
+#if HAVE_FTRUNCATE
+	have_inplace = "";
+#endif
+
 #if SUPPORT_HARD_LINKS
 	hardlinks = "";
 #endif
@@ -165,23 +183,24 @@ static void print_rsync_version(enum logcode f)
 	rprintf(f, "<http://rsync.samba.org/>\n");
 	rprintf(f, "Capabilities: %d-bit files, %ssocketpairs, "
 		"%shard links, %ssymlinks, batchfiles, \n",
-		(int) (sizeof(OFF_T) * 8),
+		(int) (sizeof (OFF_T) * 8),
 		got_socketpair, hardlinks, links);
 
 	/* Note that this field may not have type ino_t.  It depends
 	 * on the complicated interaction between largefile feature
 	 * macros. */
-	rprintf(f, "              %sIPv6, %d-bit system inums, %d-bit internal inums\n",
-		ipv6,
-		(int) (sizeof(dumstat->st_ino) * 8),
-		(int) (sizeof(INO64_T) * 8));
+	rprintf(f, "              %sinplace, %sIPv6, %d-bit system inums, %d-bit internal inums\n",
+		have_inplace, ipv6,
+		(int) (sizeof dumstat->st_ino * 8),
+		(int) (sizeof (uint64) * 8));
 #ifdef MAINTAINER_MODE
 	rprintf(f, "              panic action: \"%s\"\n",
 		get_panic_action());
 #endif
 
-#ifdef NO_INT64
-	rprintf(f, "WARNING: no 64-bit integers on this platform!\n");
+#ifdef INT64_IS_OFF_T
+	if (sizeof (int64) < 8)
+		rprintf(f, "WARNING: no 64-bit integers on this platform!\n");
 #endif
 
 	rprintf(f,
@@ -222,10 +241,12 @@ void usage(enum logcode F)
   rprintf(F,"     --backup-dir            make backups into this directory\n");
   rprintf(F,"     --suffix=SUFFIX         backup suffix (default %s w/o --backup-dir)\n",BACKUP_SUFFIX);
   rprintf(F," -u, --update                update only (don't overwrite newer files)\n");
+  rprintf(F,"     --inplace               update destination files inplace (SEE MAN PAGE)\n");
+  rprintf(F," -K, --keep-dirlinks         treat symlinked dir on receiver as dir\n");
   rprintf(F," -l, --links                 copy symlinks as symlinks\n");
-  rprintf(F," -L, --copy-links            copy the referent of symlinks\n");
-  rprintf(F,"     --copy-unsafe-links     copy links outside the source tree\n");
-  rprintf(F,"     --safe-links            ignore links outside the destination tree\n");
+  rprintf(F," -L, --copy-links            copy the referent of all symlinks\n");
+  rprintf(F,"     --copy-unsafe-links     copy the referent of \"unsafe\" symlinks\n");
+  rprintf(F,"     --safe-links            ignore \"unsafe\" symlinks\n");
   rprintf(F," -H, --hard-links            preserve hard links\n");
   rprintf(F," -p, --perms                 preserve permissions\n");
   rprintf(F," -o, --owner                 preserve owner (root only)\n");
@@ -237,24 +258,25 @@ void usage(enum logcode F)
   rprintf(F," -W, --whole-file            copy whole files, no incremental checks\n");
   rprintf(F,"     --no-whole-file         turn off --whole-file\n");
   rprintf(F," -x, --one-file-system       don't cross filesystem boundaries\n");
-  rprintf(F," -B, --block-size=SIZE       checksum blocking size (default %d)\n",BLOCK_SIZE);
+  rprintf(F," -B, --block-size=SIZE       force a fixed checksum block-size\n");
   rprintf(F," -e, --rsh=COMMAND           specify the remote shell\n");
   rprintf(F,"     --rsync-path=PATH       specify path to rsync on the remote machine\n");
   rprintf(F,"     --existing              only update files that already exist\n");
   rprintf(F,"     --ignore-existing       ignore files that already exist on receiving side\n");
   rprintf(F,"     --delete                delete files that don't exist on the sending side\n");
   rprintf(F,"     --delete-excluded       also delete excluded files on the receiving side\n");
-  rprintf(F,"     --delete-after          delete after transferring, not before\n");
-  rprintf(F,"     --ignore-errors         delete even if there are IO errors\n");
+  rprintf(F,"     --delete-after          receiver deletes after transferring, not before\n");
+  rprintf(F,"     --ignore-errors         delete even if there are I/O errors\n");
   rprintf(F,"     --max-delete=NUM        don't delete more than NUM files\n");
   rprintf(F,"     --partial               keep partially transferred files\n");
+  rprintf(F,"     --partial-dir=DIR       put a partially transferred file into DIR\n");
   rprintf(F,"     --force                 force deletion of directories even if not empty\n");
   rprintf(F,"     --numeric-ids           don't map uid/gid values by user/group name\n");
-  rprintf(F,"     --timeout=TIME          set IO timeout in seconds\n");
+  rprintf(F,"     --timeout=TIME          set I/O timeout in seconds\n");
   rprintf(F," -I, --ignore-times          turn off mod time & file size quick check\n");
   rprintf(F,"     --size-only             ignore mod time for quick check (use size)\n");
   rprintf(F,"     --modify-window=NUM     compare mod times with reduced accuracy\n");
-  rprintf(F," -T  --temp-dir=DIR          create temporary files in directory DIR\n");
+  rprintf(F," -T, --temp-dir=DIR          create temporary files in directory DIR\n");
   rprintf(F,"     --compare-dest=DIR      also compare destination files relative to DIR\n");
   rprintf(F,"     --link-dest=DIR         create hardlinks to DIR for unchanged files\n");
   rprintf(F," -P                          equivalent to --partial --progress\n");
@@ -265,27 +287,28 @@ void usage(enum logcode F)
   rprintf(F,"     --include=PATTERN       don't exclude files matching PATTERN\n");
   rprintf(F,"     --include-from=FILE     don't exclude patterns listed in FILE\n");
   rprintf(F,"     --files-from=FILE       read FILE for list of source-file names\n");
-  rprintf(F," -0  --from0                 all *-from file lists are delimited by nulls\n");
+  rprintf(F," -0, --from0                 all *-from file lists are delimited by nulls\n");
   rprintf(F,"     --version               print version number\n");
-  rprintf(F,"     --daemon                run as a rsync daemon\n");
+  rprintf(F,"     --daemon                run as an rsync daemon\n");
   rprintf(F,"     --no-detach             do not detach from the parent\n");
   rprintf(F,"     --address=ADDRESS       bind to the specified address\n");
   rprintf(F,"     --config=FILE           specify alternate rsyncd.conf file\n");
   rprintf(F,"     --port=PORT             specify alternate rsyncd port number\n");
-  rprintf(F,"     --blocking-io           use blocking IO for the remote shell\n");
+  rprintf(F,"     --blocking-io           use blocking I/O for the remote shell\n");
   rprintf(F,"     --no-blocking-io        turn off --blocking-io\n");
   rprintf(F,"     --stats                 give some file transfer stats\n");
   rprintf(F,"     --progress              show progress during transfer\n");
   rprintf(F,"     --log-format=FORMAT     log file transfers using specified format\n");
   rprintf(F,"     --password-file=FILE    get password from FILE\n");
   rprintf(F,"     --bwlimit=KBPS          limit I/O bandwidth, KBytes per second\n");
-  rprintf(F,"     --write-batch=PREFIX    write batch fileset starting with PREFIX\n");
-  rprintf(F,"     --read-batch=PREFIX     read batch fileset starting with PREFIX\n");
-  rprintf(F," -h, --help                  show this help screen\n");
+  rprintf(F,"     --write-batch=FILE      write a batch to FILE\n");
+  rprintf(F,"     --read-batch=FILE       read a batch from FILE\n");
+  rprintf(F,"     --checksum-seed=NUM     set block/file checksum seed\n");
 #ifdef INET6
-  rprintf(F," -4                          prefer IPv4\n");
-  rprintf(F," -6                          prefer IPv6\n");
+  rprintf(F," -4, --ipv4                  prefer IPv4\n");
+  rprintf(F," -6, --ipv6                  prefer IPv6\n");
 #endif
+  rprintf(F," -h, --help                  show this help screen\n");
 
   rprintf(F,"\n");
 
@@ -296,14 +319,15 @@ void usage(enum logcode F)
 enum {OPT_VERSION = 1000, OPT_SENDER, OPT_EXCLUDE, OPT_EXCLUDE_FROM,
       OPT_DELETE_AFTER, OPT_DELETE_EXCLUDED, OPT_LINK_DEST,
       OPT_INCLUDE, OPT_INCLUDE_FROM, OPT_MODIFY_WINDOW,
-      OPT_READ_BATCH, OPT_WRITE_BATCH};
+      OPT_READ_BATCH, OPT_WRITE_BATCH, OPT_TIMEOUT,
+      OPT_REFUSED_BASE = 9000};
 
 static struct poptOption long_options[] = {
   /* longName, shortName, argInfo, argPtr, value, descrip, argDesc */
   {"version",          0,  POPT_ARG_NONE,   0,              OPT_VERSION, 0, 0},
   {"suffix",           0,  POPT_ARG_STRING, &backup_suffix, 0, 0, 0 },
-  {"rsync-path",       0,  POPT_ARG_STRING, &rsync_path,	0, 0, 0 },
-  {"password-file",    0,  POPT_ARG_STRING, &password_file,	0, 0, 0 },
+  {"rsync-path",       0,  POPT_ARG_STRING, &rsync_path, 0, 0, 0 },
+  {"password-file",    0,  POPT_ARG_STRING, &password_file, 0, 0, 0 },
   {"ignore-times",    'I', POPT_ARG_NONE,   &ignore_times, 0, 0, 0 },
   {"size-only",        0,  POPT_ARG_NONE,   &size_only, 0, 0, 0 },
   {"modify-window",    0,  POPT_ARG_INT,    &modify_window, OPT_MODIFY_WINDOW, 0, 0 },
@@ -326,6 +350,8 @@ static struct poptOption long_options[] = {
   {"sparse",          'S', POPT_ARG_NONE,   &sparse_files, 0, 0, 0 },
   {"cvs-exclude",     'C', POPT_ARG_NONE,   &cvs_exclude, 0, 0, 0 },
   {"update",          'u', POPT_ARG_NONE,   &update_only, 0, 0, 0 },
+  {"inplace",          0,  POPT_ARG_NONE,   &inplace, 0, 0, 0 },
+  {"keep-dirlinks",   'K', POPT_ARG_NONE,   &keep_dirlinks, 0, 0, 0 },
   {"links",           'l', POPT_ARG_NONE,   &preserve_links, 0, 0, 0 },
   {"copy-links",      'L', POPT_ARG_NONE,   &copy_links, 0, 0, 0 },
   {"whole-file",      'W', POPT_ARG_VAL,    &whole_file, 1, 0, 0 },
@@ -348,17 +374,18 @@ static struct poptOption long_options[] = {
   {"rsh",             'e', POPT_ARG_STRING, &shell_cmd, 0, 0, 0 },
   {"block-size",      'B', POPT_ARG_INT,    &block_size, 0, 0, 0 },
   {"max-delete",       0,  POPT_ARG_INT,    &max_delete, 0, 0, 0 },
-  {"timeout",          0,  POPT_ARG_INT,    &io_timeout, 0, 0, 0 },
+  {"timeout",          0,  POPT_ARG_INT,    &io_timeout, OPT_TIMEOUT, 0, 0 },
   {"temp-dir",        'T', POPT_ARG_STRING, &tmpdir, 0, 0, 0 },
   {"compare-dest",     0,  POPT_ARG_STRING, &compare_dest, 0, 0, 0 },
-  {"link-dest",        0,  POPT_ARG_STRING, 0,              OPT_LINK_DEST, 0, 0 },
+  {"link-dest",        0,  POPT_ARG_STRING, &compare_dest,  OPT_LINK_DEST, 0, 0 },
   /* TODO: Should this take an optional int giving the compression level? */
   {"compress",        'z', POPT_ARG_NONE,   &do_compression, 0, 0, 0 },
-  {"daemon",           0,  POPT_ARG_NONE,   &am_daemon, 0, 0, 0 },
+  {"daemon",           0,  POPT_ARG_NONE,   &daemon_opt, 0, 0, 0 },
   {"no-detach",        0,  POPT_ARG_NONE,   &no_detach, 0, 0, 0 },
   {"stats",            0,  POPT_ARG_NONE,   &do_stats, 0, 0, 0 },
   {"progress",         0,  POPT_ARG_NONE,   &do_progress, 0, 0, 0 },
   {"partial",          0,  POPT_ARG_NONE,   &keep_partial, 0, 0, 0 },
+  {"partial-dir",      0,  POPT_ARG_STRING, &partial_dir, 0, 0, 0 },
   {"ignore-errors",    0,  POPT_ARG_NONE,   &ignore_errors, 0, 0, 0 },
   {"blocking-io",      0,  POPT_ARG_VAL,    &blocking_io, 1, 0, 0 },
   {"no-blocking-io",   0,  POPT_ARG_VAL,    &blocking_io, 0, 0, 0 },
@@ -370,21 +397,22 @@ static struct poptOption long_options[] = {
   {"address",          0,  POPT_ARG_STRING, &bind_address, 0, 0, 0 },
   {"backup-dir",       0,  POPT_ARG_STRING, &backup_dir, 0, 0, 0 },
   {"hard-links",      'H', POPT_ARG_NONE,   &preserve_hard_links, 0, 0, 0 },
-  {"read-batch",       0,  POPT_ARG_STRING, &batch_prefix,  OPT_READ_BATCH, 0, 0 },
-  {"write-batch",      0,  POPT_ARG_STRING, &batch_prefix,  OPT_WRITE_BATCH, 0, 0 },
+  {"read-batch",       0,  POPT_ARG_STRING, &batch_name,  OPT_READ_BATCH, 0, 0 },
+  {"write-batch",      0,  POPT_ARG_STRING, &batch_name,  OPT_WRITE_BATCH, 0, 0 },
   {"files-from",       0,  POPT_ARG_STRING, &files_from, 0, 0, 0 },
   {"from0",           '0', POPT_ARG_NONE,   &eol_nulls, 0, 0, 0},
   {"no-implied-dirs",  0,  POPT_ARG_VAL,    &implied_dirs, 0, 0, 0 },
   {"protocol",         0,  POPT_ARG_INT,    &protocol_version, 0, 0, 0 },
+  {"checksum-seed",    0,  POPT_ARG_INT,    &checksum_seed, 0, 0, 0 },
 #ifdef INET6
-  {0,		      '4', POPT_ARG_VAL,    &default_af_hint, AF_INET, 0, 0 },
-  {0,		      '6', POPT_ARG_VAL,    &default_af_hint, AF_INET6, 0, 0 },
+  {"ipv4",            '4', POPT_ARG_VAL,    &default_af_hint, AF_INET, 0, 0 },
+  {"ipv6",            '6', POPT_ARG_VAL,    &default_af_hint, AF_INET6, 0, 0 },
 #endif
   {0,0,0,0, 0, 0, 0}
 };
 
 
-static char err_buf[100];
+static char err_buf[200];
 
 
 /**
@@ -394,46 +422,59 @@ static char err_buf[100];
  **/
 void option_error(void)
 {
-	if (err_buf[0]) {
-		rprintf(FLOG, "%s", err_buf);
-		rprintf(FERROR, RSYNC_NAME ": %s", err_buf);
-	} else {
-		rprintf (FERROR, "Error parsing options: "
-			 "option may be supported on client but not on server?\n");
-		rprintf (FERROR, RSYNC_NAME ": Error parsing options: "
-			 "option may be supported on client but not on server?\n");
+	if (!err_buf[0]) {
+		strcpy(err_buf, "Error parsing options: "
+		    "option may be supported on client but not on server?\n");
 	}
+
+	rprintf(FERROR, RSYNC_NAME ": %s", err_buf);
 }
 
 
 /**
- * Check to see if we should refuse this option
+ * Tweak the option table to disable all options that the rsyncd.conf
+ * file has told us to refuse.
  **/
-static int check_refuse_options(char *ref, int opt)
+static void set_refuse_options(char *bp)
 {
-	int i, len;
-	char *p;
-	const char *name;
+	struct poptOption *op;
+	char *cp, shortname[2];
+	int is_wild, found_match;
 
-	for (i=0; long_options[i].longName; i++) {
-		if (long_options[i].val == opt) break;
-	}
+	shortname[1] = '\0';
 
-	if (!long_options[i].longName) return 0;
-
-	name = long_options[i].longName;
-	len = strlen(name);
-
-	while ((p = strstr(ref,name))) {
-		if ((p==ref || p[-1]==' ') &&
-		    (p[len] == ' ' || p[len] == 0)) {
-			snprintf(err_buf,sizeof(err_buf),
-				 "The '%s' option is not supported by this server\n", name);
-			return 1;
+	while (1) {
+		while (*bp == ' ') bp++;
+		if (!*bp)
+			break;
+		if ((cp = strchr(bp, ' ')) != NULL)
+			*cp= '\0';
+		/* If they specify "delete", reject all delete options. */
+		if (strcmp(bp, "delete") == 0)
+			bp = "delete*";
+		is_wild = strpbrk(bp, "*?[") != NULL;
+		found_match = 0;
+		for (op = long_options; ; op++) {
+			*shortname = op->shortName;
+			if (!op->longName && !*shortname)
+				break;
+			if ((op->longName && wildmatch(bp, op->longName))
+			    || (*shortname && wildmatch(bp, shortname))) {
+				op->val = (op - long_options) + OPT_REFUSED_BASE;
+				found_match = 1;
+				if (!is_wild)
+					break;
+			}
 		}
-		ref += len;
+		if (!found_match) {
+			rprintf(FLOG, "No match for refuse-options string \"%s\"\n",
+				bp);
+		}
+		if (!cp)
+			break;
+		*cp = ' ';
+		bp = cp + 1;
 	}
-	return 0;
 }
 
 
@@ -460,8 +501,12 @@ int parse_arguments(int *argc, const char ***argv, int frommain)
 {
 	int opt;
 	char *ref = lp_refuse_options(module_id);
+	const char *arg;
 	poptContext pc;
 
+	if (ref && *ref)
+		set_refuse_options(ref);
+
 	/* TODO: Call poptReadDefaultConfig; handle errors. */
 
 	/* The context leaks in case of an error, but if there's a
@@ -469,10 +514,6 @@ int parse_arguments(int *argc, const char ***argv, int frommain)
 	pc = poptGetContext(RSYNC_NAME, *argc, *argv, long_options, 0);
 
 	while ((opt = poptGetNextOpt(pc)) != -1) {
-		if (ref) {
-			if (check_refuse_options(ref, opt)) return 0;
-		}
-
 		/* most options are handled automatically by popt;
 		 * only special cases are returned and listed here. */
 
@@ -499,53 +540,41 @@ int parse_arguments(int *argc, const char ***argv, int frommain)
 			break;
 
 		case OPT_EXCLUDE:
-			add_exclude(&exclude_list, poptGetOptArg(pc),
-				    ADD_EXCLUDE);
+			add_exclude(&exclude_list, poptGetOptArg(pc), 0);
 			break;
 
 		case OPT_INCLUDE:
 			add_exclude(&exclude_list, poptGetOptArg(pc),
-				    ADD_INCLUDE);
+				    XFLG_DEF_INCLUDE);
 			break;
 
 		case OPT_EXCLUDE_FROM:
-			add_exclude_file(&exclude_list, poptGetOptArg(pc),
-					 MISSING_FATAL, ADD_EXCLUDE);
-			break;
-
 		case OPT_INCLUDE_FROM:
-			add_exclude_file(&exclude_list, poptGetOptArg(pc),
-					 MISSING_FATAL, ADD_INCLUDE);
+			arg = poptGetOptArg(pc);
+			if (sanitize_paths)
+				arg = sanitize_path(NULL, arg, NULL, 0);
+			if (server_exclude_list.head) {
+				char *cp = (char *)arg;
+				clean_fname(cp, 1);
+				if (check_exclude(&server_exclude_list, cp, 0) < 0)
+					goto options_rejected;
+			}
+			add_exclude_file(&exclude_list, arg, XFLG_FATAL_ERRORS
+					 | (opt == OPT_INCLUDE_FROM
+					  ? XFLG_DEF_INCLUDE : 0));
 			break;
 
 		case 'h':
 			usage(FINFO);
 			exit_cleanup(0);
 
-		case 'H':
-#if SUPPORT_HARD_LINKS
-			preserve_hard_links=1;
-#else
-			/* FIXME: Don't say "server" if this is
-			 * happening on the client. */
-			/* FIXME: Why do we have the duplicated
-			 * rprintf?  Everybody who gets this message
-			 * ought to send it to the client and also to
-			 * the logs. */
-			snprintf(err_buf, sizeof err_buf,
-				 "hard links are not supported on this %s\n",
-				 am_server ? "server" : "client");
-			rprintf(FERROR, "ERROR: %s", err_buf);
-			return 0;
-#endif /* SUPPORT_HARD_LINKS */
-			break;
-
 		case 'v':
 			verbose++;
 			break;
 
 		case 'q':
-			if (frommain) quiet++;
+			if (frommain)
+				quiet++;
 			break;
 
 		case OPT_SENDER:
@@ -562,51 +591,112 @@ int parse_arguments(int *argc, const char ***argv, int frommain)
 			break;
 
 		case OPT_WRITE_BATCH:
-			/* popt stores the filename in batch_prefix for us */
+			/* batch_name is already set */
 			write_batch = 1;
 			break;
 
 		case OPT_READ_BATCH:
-			/* popt stores the filename in batch_prefix for us */
+			/* batch_name is already set */
 			read_batch = 1;
 			break;
 
+		case OPT_TIMEOUT:
+			if (io_timeout && io_timeout < select_timeout)
+				select_timeout = io_timeout;
+			break;
+
 		case OPT_LINK_DEST:
 #if HAVE_LINK
-			compare_dest = (char *)poptGetOptArg(pc);
 			link_dest = 1;
 			break;
 #else
 			snprintf(err_buf, sizeof err_buf,
 				 "hard links are not supported on this %s\n",
 				 am_server ? "server" : "client");
-			rprintf(FERROR, "ERROR: %s", err_buf);
 			return 0;
 #endif
 
-
 		default:
-			/* FIXME: If --daemon is specified, then errors for later
-			 * parameters seem to disappear. */
-			snprintf(err_buf, sizeof(err_buf),
-				 "%s%s: %s\n",
-				 am_server ? "on remote machine: " : "",
-				 poptBadOption(pc, POPT_BADOPTION_NOALIAS),
-				 poptStrerror(opt));
+			/* A large opt value means that set_refuse_options()
+			 * turned this option off (opt-BASE is its index). */
+			if (opt >= OPT_REFUSED_BASE) {
+				struct poptOption *op =
+				    &long_options[opt-OPT_REFUSED_BASE];
+				int n = snprintf(err_buf, sizeof err_buf,
+				    "The server is configured to refuse --%s\n",
+				    op->longName) - 1;
+				if (op->shortName) {
+					snprintf(err_buf+n, sizeof err_buf-n,
+					    " (-%c)\n", op->shortName);
+				}
+			} else {
+				snprintf(err_buf, sizeof err_buf,
+				    "%s%s: %s\n",
+				    am_server ? "on remote machine: " : "",
+				    poptBadOption(pc, POPT_BADOPTION_NOALIAS),
+				    poptStrerror(opt));
+			}
 			return 0;
 		}
 	}
 
+#if !SUPPORT_LINKS
+	if (preserve_links && !am_sender) {
+		snprintf(err_buf, sizeof err_buf,
+			 "symlinks are not supported on this %s\n",
+			 am_server ? "server" : "client");
+		return 0;
+	}
+#endif
+
+#if !SUPPORT_HARD_LINKS
+	if (preserve_hard_links) {
+		snprintf(err_buf, sizeof err_buf,
+			 "hard links are not supported on this %s\n",
+			 am_server ? "server" : "client");
+		return 0;
+	}
+#endif
+
 	if (write_batch && read_batch) {
-		rprintf(FERROR,
-			"write-batch and read-batch can not be used together\n");
-		exit_cleanup(RERR_SYNTAX);
+		snprintf(err_buf, sizeof err_buf,
+			"--write-batch and --read-batch can not be used together\n");
+		return 0;
+	}
+	if (write_batch || read_batch) {
+		if (dry_run) {
+			snprintf(err_buf, sizeof err_buf,
+				"--%s-batch cannot be used with --dry_run (-n)\n",
+				write_batch ? "write" : "read");
+			return 0;
+		}
+		if (am_server) {
+			rprintf(FINFO,
+				"ignoring --%s-batch option sent to server\n",
+				write_batch ? "write" : "read");
+			/* We don't actually exit_cleanup(), so that we can
+			 * still service older version clients that still send
+			 * batch args to server. */
+			read_batch = write_batch = 0;
+			batch_name = NULL;
+		}
+	}
+	if (read_batch && files_from) {
+		snprintf(err_buf, sizeof err_buf,
+			"--read-batch cannot be used with --files-from\n");
+		return 0;
+	}
+	if (batch_name && strlen(batch_name) > MAX_BATCH_NAME_LEN) {
+		snprintf(err_buf, sizeof err_buf,
+			"the batch-file name must be %d characters or less.\n",
+			MAX_BATCH_NAME_LEN);
+		return 0;
 	}
 
-	if (do_compression && (write_batch || read_batch)) {
-		rprintf(FERROR,
-			"compress can not be used with write-batch or read-batch\n");
-		exit_cleanup(RERR_SYNTAX);
+	if (tmpdir && strlen(tmpdir) >= MAXPATHLEN - 10) {
+		snprintf(err_buf, sizeof err_buf,
+			 "the --temp-dir path is WAY too long.\n");
+		return 0;
 	}
 
 	if (archive_mode) {
@@ -625,34 +715,141 @@ int parse_arguments(int *argc, const char ***argv, int frommain)
 	if (relative_paths < 0)
 		relative_paths = files_from? 1 : 0;
 
-	if (!backup_suffix)
-		backup_suffix = backup_dir? "" : BACKUP_SUFFIX;
-	backup_suffix_len = strlen(backup_suffix);
-	if (strchr(backup_suffix, '/') != NULL) {
-		rprintf(FERROR, "--suffix cannot contain slashes: %s\n",
-			backup_suffix);
-		exit_cleanup(RERR_SYNTAX);
-	}
-	if (backup_dir)
-		backup_dir_len = strlen(backup_dir);
-	else if (!backup_suffix_len) {
-		rprintf(FERROR,
-			"--suffix cannot be a null string without --backup-dir\n");
-		exit_cleanup(RERR_SYNTAX);
-	}
-
-	if (do_progress && !verbose)
-		verbose = 1;
-
 	*argv = poptGetArgs(pc);
 	if (*argv)
 		*argc = count_args(*argv);
 	else
 		*argc = 0;
 
+	if (sanitize_paths) {
+		int i;
+		for (i = *argc; i-- > 0; )
+			(*argv)[i] = sanitize_path(NULL, (*argv)[i], "", 0);
+		if (tmpdir)
+			tmpdir = sanitize_path(NULL, tmpdir, NULL, 0);
+		if (partial_dir)
+			partial_dir = sanitize_path(NULL, partial_dir, NULL, 0);
+		if (compare_dest)
+			compare_dest = sanitize_path(NULL, compare_dest, NULL, 0);
+		if (backup_dir)
+			backup_dir = sanitize_path(NULL, backup_dir, NULL, 0);
+		if (files_from)
+			files_from = sanitize_path(NULL, files_from, NULL, 0);
+	}
+	if (server_exclude_list.head && !am_sender) {
+		struct exclude_list_struct *elp = &server_exclude_list;
+		if (tmpdir) {
+			clean_fname(tmpdir, 1);
+			if (check_exclude(elp, tmpdir, 1) < 0)
+				goto options_rejected;
+		}
+		if (partial_dir) {
+			clean_fname(partial_dir, 1);
+			if (check_exclude(elp, partial_dir, 1) < 0)
+				goto options_rejected;
+		}
+		if (compare_dest) {
+			clean_fname(compare_dest, 1);
+			if (check_exclude(elp, compare_dest, 1) < 0)
+				goto options_rejected;
+		}
+		if (backup_dir) {
+			clean_fname(backup_dir, 1);
+			if (check_exclude(elp, backup_dir, 1) < 0)
+				goto options_rejected;
+		}
+	}
+	if (server_exclude_list.head && files_from) {
+		clean_fname(files_from, 1);
+		if (check_exclude(&server_exclude_list, files_from, 0) < 0) {
+		    options_rejected:
+			snprintf(err_buf, sizeof err_buf,
+			    "Your options have been rejected by the server.\n");
+			return 0;
+		}
+	}
+
+	if (daemon_opt) {
+		daemon_opt = 0;
+		am_daemon = 1;
+		return 1;
+	}
+
+	if (!backup_suffix)
+		backup_suffix = backup_dir ? "" : BACKUP_SUFFIX;
+	backup_suffix_len = strlen(backup_suffix);
+	if (strchr(backup_suffix, '/') != NULL) {
+		snprintf(err_buf, sizeof err_buf,
+			"--suffix cannot contain slashes: %s\n",
+			backup_suffix);
+		return 0;
+	}
+	if (backup_dir) {
+		backup_dir_len = strlcpy(backup_dir_buf, backup_dir, sizeof backup_dir_buf);
+		backup_dir_remainder = sizeof backup_dir_buf - backup_dir_len;
+		if (backup_dir_remainder < 32) {
+			snprintf(err_buf, sizeof err_buf,
+				"the --backup-dir path is WAY too long.\n");
+			return 0;
+		}
+		if (backup_dir_buf[backup_dir_len - 1] != '/') {
+			backup_dir_buf[backup_dir_len++] = '/';
+			backup_dir_buf[backup_dir_len] = '\0';
+		}
+		if (verbose > 1 && !am_sender)
+			rprintf(FINFO, "backup_dir is %s\n", backup_dir_buf);
+	} else if (!backup_suffix_len && (!am_server || !am_sender)) {
+		snprintf(err_buf, sizeof err_buf,
+			"--suffix cannot be a null string without --backup-dir\n");
+		return 0;
+	}
+
+	if (do_progress && !verbose)
+		verbose = 1;
+
+	if (bwlimit) {
+		bwlimit_writemax = (size_t)bwlimit * 128;
+		if (bwlimit_writemax < 512)
+			bwlimit_writemax = 512;
+	}
+
+	if (inplace) {
+#if HAVE_FTRUNCATE
+		if (partial_dir) {
+			snprintf(err_buf, sizeof err_buf,
+				 "--inplace cannot be used with --partial-dir\n");
+			return 0;
+		}
+		keep_partial = 0;
+#else
+		snprintf(err_buf, sizeof err_buf,
+			 "--inplace is not supported on this %s\n",
+			 am_server ? "server" : "client");
+		return 0;
+#endif
+		if (compare_dest) {
+			snprintf(err_buf, sizeof err_buf,
+				 "--inplace does not yet work with %s\n",
+				 link_dest ? "--link-dest" : "--compare-dest");
+			return 0;
+		}
+	} else {
+		if (keep_partial && !partial_dir)
+			partial_dir = getenv("RSYNC_PARTIAL_DIR");
+		if (partial_dir) {
+			if (!*partial_dir || strcmp(partial_dir, ".") == 0)
+				partial_dir = NULL;
+			else if (*partial_dir != '/') {
+				add_exclude(&exclude_list, partial_dir,
+					    XFLG_DIRECTORY);
+			}
+			keep_partial = 1;
+		}
+	}
+
 	if (files_from) {
 		char *colon;
-		if (*argc != 2) {
+		if (*argc > 2 || (!am_daemon && *argc == 1)) {
 			usage(FERROR);
 			exit_cleanup(RERR_SYNTAX);
 		}
@@ -668,19 +865,17 @@ int parse_arguments(int *argc, const char ***argv, int frommain)
 			}
 			remote_filesfrom_file = colon+1 + (colon[1] == ':');
 			if (strcmp(remote_filesfrom_file, "-") == 0) {
-				rprintf(FERROR, "Invalid --files-from remote filename\n");
-				exit_cleanup(RERR_SYNTAX);
+				snprintf(err_buf, sizeof err_buf,
+					"Invalid --files-from remote filename\n");
+				return 0;
 			}
 		} else {
-			extern int sanitize_paths;
-			if (sanitize_paths)
-				sanitize_path(strdup(files_from), NULL);
 			filesfrom_fd = open(files_from, O_RDONLY|O_BINARY);
 			if (filesfrom_fd < 0) {
-				rsyserr(FERROR, errno,
-					"failed to open files-from file %s",
-					files_from);
-				exit_cleanup(RERR_FILEIO);
+				snprintf(err_buf, sizeof err_buf,
+					"failed to open files-from file %s: %s\n",
+					files_from, strerror(errno));
+				return 0;
 			}
 		}
 	}
@@ -701,13 +896,7 @@ void server_options(char **args,int *argc)
 {
 	int ac = *argc;
 	static char argstr[50];
-	static char bsize[30];
-	static char iotime[30];
-	static char mdelete[30];
-	static char mwindow[30];
-	static char bw[50];
-	/* Leave room for ``--(write|read)-batch='' */
-	static char fext[MAXPATHLEN + 15];
+	char *arg;
 
 	int i, x;
 
@@ -728,7 +917,7 @@ void server_options(char **args,int *argc)
 
 	x = 1;
 	argstr[0] = '-';
-	for (i=0;i<verbose;i++)
+	for (i = 0; i < verbose; i++)
 		argstr[x++] = 'v';
 
 	/* the -q option is intentionally left out */
@@ -742,6 +931,8 @@ void server_options(char **args,int *argc)
 		argstr[x++] = 'l';
 	if (copy_links)
 		argstr[x++] = 'L';
+	if (keep_dirlinks && am_sender)
+		argstr[x++] = 'K';
 
 	if (whole_file > 0)
 		argstr[x++] = 'W';
@@ -787,37 +978,31 @@ void server_options(char **args,int *argc)
 
 	argstr[x] = 0;
 
-	if (x != 1) args[ac++] = argstr;
+	if (x != 1)
+		args[ac++] = argstr;
 
 	if (block_size) {
-		snprintf(bsize,sizeof(bsize),"-B%d",block_size);
-		args[ac++] = bsize;
+		if (asprintf(&arg, "-B%u", block_size) < 0)
+			goto oom;
+		args[ac++] = arg;
 	}
 
 	if (max_delete && am_sender) {
-		snprintf(mdelete,sizeof(mdelete),"--max-delete=%d",max_delete);
-		args[ac++] = mdelete;
-	}
-
-	if (batch_prefix != NULL) {
-		char *fmt = "";
-		if (write_batch)
-			fmt = "--write-batch=%s";
-		else
-		if (read_batch)
-			fmt = "--read-batch=%s";
-		snprintf(fext,sizeof(fext),fmt,batch_prefix);
-		args[ac++] = fext;
+		if (asprintf(&arg, "--max-delete=%d", max_delete) < 0)
+			goto oom;
+		args[ac++] = arg;
 	}
 
 	if (io_timeout) {
-		snprintf(iotime,sizeof(iotime),"--timeout=%d",io_timeout);
-		args[ac++] = iotime;
+		if (asprintf(&arg, "--timeout=%d", io_timeout) < 0)
+			goto oom;
+		args[ac++] = arg;
 	}
 
 	if (bwlimit) {
-		snprintf(bw,sizeof(bw),"--bwlimit=%d",bwlimit);
-		args[ac++] = bw;
+		if (asprintf(&arg, "--bwlimit=%d", bwlimit) < 0)
+			goto oom;
+		args[ac++] = arg;
 	}
 
 	if (backup_dir) {
@@ -826,39 +1011,47 @@ void server_options(char **args,int *argc)
 	}
 
 	/* Only send --suffix if it specifies a non-default value. */
-	if (strcmp(backup_suffix, backup_dir? "" : BACKUP_SUFFIX) != 0) {
-		char *s = new_array(char, 9+backup_suffix_len+1);
-		if (!s)
-			out_of_memory("server_options");
+	if (strcmp(backup_suffix, backup_dir ? "" : BACKUP_SUFFIX) != 0) {
 		/* We use the following syntax to avoid weirdness with '~'. */
-		sprintf(s, "--suffix=%s", backup_suffix);
-		args[ac++] = s;
+		if (asprintf(&arg, "--suffix=%s", backup_suffix) < 0)
+			goto oom;
+		args[ac++] = arg;
 	}
 
-	if (delete_mode && !delete_excluded)
-		args[ac++] = "--delete";
+	if (am_sender) {
+		if (delete_excluded)
+			args[ac++] = "--delete-excluded";
+		else if (delete_mode)
+			args[ac++] = "--delete";
 
-	if (delete_excluded)
-		args[ac++] = "--delete-excluded";
+		if (delete_after)
+			args[ac++] = "--delete-after";
+
+		if (force_delete)
+			args[ac++] = "--force";
+	}
 
 	if (size_only)
 		args[ac++] = "--size-only";
 
 	if (modify_window_set) {
-		snprintf(mwindow,sizeof(mwindow),"--modify-window=%d",
-			 modify_window);
-		args[ac++] = mwindow;
+		if (asprintf(&arg, "--modify-window=%d", modify_window) < 0)
+			goto oom;
+		args[ac++] = arg;
 	}
 
-	if (keep_partial)
+	if (checksum_seed) {
+		if (asprintf(&arg, "--checksum-seed=%d", checksum_seed) < 0)
+			goto oom;
+		args[ac++] = arg;
+	}
+
+	if (partial_dir && am_sender) {
+		args[ac++] = "--partial-dir";
+		args[ac++] = partial_dir;
+	} else if (keep_partial)
 		args[ac++] = "--partial";
 
-	if (force_delete)
-		args[ac++] = "--force";
-
-	if (delete_after)
-		args[ac++] = "--delete-after";
-
 	if (ignore_errors)
 		args[ac++] = "--ignore-errors";
 
@@ -877,6 +1070,9 @@ void server_options(char **args,int *argc)
 	if (opt_ignore_existing && am_sender)
 		args[ac++] = "--ignore-existing";
 
+	if (inplace)
+		args[ac++] = "--inplace";
+
 	if (tmpdir) {
 		args[ac++] = "--temp-dir";
 		args[ac++] = tmpdir;
@@ -904,6 +1100,10 @@ void server_options(char **args,int *argc)
 	}
 
 	*argc = ac;
+	return;
+
+    oom:
+	out_of_memory("server_options");
 }
 
 /**
@@ -915,12 +1115,14 @@ char *find_colon(char *s)
 	char *p, *p2;
 
 	p = strchr(s,':');
-	if (!p) return NULL;
+	if (!p)
+		return NULL;
 
 	/* now check to see if there is a / in the string before the : - if there is then
 	   discard the colon on the assumption that the : is part of a filename */
 	p2 = strchr(s,'/');
-	if (p2 && p2 < p) return NULL;
+	if (p2 && p2 < p)
+		return NULL;
 
 	return p;
 }
