@@ -22,19 +22,22 @@
 
 extern int verbose;
 extern int recurse;
-extern int delete_mode;
 extern int delete_after;
 extern int max_delete;
 extern int csum_length;
 extern struct stats stats;
 extern int dry_run;
+extern int read_batch;
+extern int batch_gen_fd;
 extern int am_server;
 extern int relative_paths;
+extern int keep_dirlinks;
 extern int preserve_hard_links;
 extern int preserve_perms;
 extern int cvs_exclude;
 extern int io_error;
 extern char *tmpdir;
+extern char *partial_dir;
 extern char *compare_dest;
 extern int make_backups;
 extern int do_progress;
@@ -45,24 +48,31 @@ extern int cleanup_got_literal;
 extern int module_id;
 extern int ignore_errors;
 extern int orig_umask;
+extern int keep_partial;
+extern int checksum_seed;
+extern int inplace;
+
+extern struct exclude_list_struct server_exclude_list;
+
 
 static void delete_one(char *fn, int is_dir)
 {
 	if (!is_dir) {
 		if (robust_unlink(fn) != 0) {
-			rprintf(FERROR, "delete_one: unlink %s failed: %s\n",
-				full_fname(fn), strerror(errno));
-		} else if (verbose) {
-			rprintf(FINFO, "deleting %s\n", fn);
-		}
+			rsyserr(FERROR, errno, "delete_one: unlink %s failed",
+				full_fname(fn));
+		} else if (verbose)
+			rprintf(FINFO, "deleting %s\n", safe_fname(fn));
 	} else {
 		if (do_rmdir(fn) != 0) {
 			if (errno != ENOTEMPTY && errno != EEXIST) {
-				rprintf(FERROR, "delete_one: rmdir %s failed: %s\n",
-					full_fname(fn), strerror(errno));
+				rsyserr(FERROR, errno,
+					"delete_one: rmdir %s failed",
+					full_fname(fn));
 			}
 		} else if (verbose) {
-			rprintf(FINFO, "deleting directory %s\n", fn);
+			rprintf(FINFO, "deleting directory %s\n",
+				safe_fname(fn));
 		}
 	}
 }
@@ -103,7 +113,7 @@ void delete_files(struct file_list *flist)
 			continue;
 
 		if (verbose > 1)
-			rprintf(FINFO, "deleting in %s\n", fbuf);
+			rprintf(FINFO, "deleting in %s\n", safe_fname(fbuf));
 
 		for (i = local_file_list->count-1; i >= 0; i--) {
 			if (max_delete && deletion_count > max_delete)
@@ -113,9 +123,11 @@ void delete_files(struct file_list *flist)
 			if (flist_find(flist,local_file_list->files[i]) < 0) {
 				char *f = f_name(local_file_list->files[i]);
 				if (make_backups && (backup_dir || !is_backup_file(f))) {
-					(void) make_backup(f);
-					if (verbose)
-						rprintf(FINFO, "deleting %s\n", f);
+					make_backup(f);
+					if (verbose) {
+						rprintf(FINFO, "deleting %s\n",
+							safe_fname(f));
+					}
 				} else {
 					int mode = local_file_list->files[i]->mode;
 					delete_one(f, S_ISDIR(mode) != 0);
@@ -189,22 +201,33 @@ static int get_tmpname(char *fnametmp, char *fname)
 }
 
 
-static int receive_data(int f_in,struct map_struct *mapbuf,int fd,char *fname,
-			OFF_T total_size)
+static int receive_data(int f_in, char *fname_r, int fd_r, OFF_T size_r,
+			char *fname, int fd, OFF_T total_size)
 {
-	int i;
+	static char file_sum1[MD4_SUM_LENGTH];
+	static char file_sum2[MD4_SUM_LENGTH];
+	struct map_struct *mapbuf;
 	struct sum_struct sum;
 	unsigned int len;
 	OFF_T offset = 0;
 	OFF_T offset2;
 	char *data;
-	static char file_sum1[MD4_SUM_LENGTH];
-	static char file_sum2[MD4_SUM_LENGTH];
-	char *map=NULL;
+	int i;
+	char *map = NULL;
 
 	read_sum_head(f_in, &sum);
 
-	sum_init();
+	if (fd_r >= 0 && size_r > 0) {
+		OFF_T map_size = MAX(sum.blength * 2, 16*1024);
+		mapbuf = map_file(fd_r, size_r, map_size, sum.blength);
+		if (verbose > 2) {
+			rprintf(FINFO, "recv mapped %s of size %.0f\n",
+				safe_fname(fname_r), (double)size_r);
+		}
+	} else
+		mapbuf = NULL;
+
+	sum_init(checksum_seed);
 
 	while ((i = recv_token(f_in, &data)) != 0) {
 		if (do_progress)
@@ -221,11 +244,8 @@ static int receive_data(int f_in,struct map_struct *mapbuf,int fd,char *fname,
 
 			sum_update(data,i);
 
-			if (fd != -1 && write_file(fd,data,i) != i) {
-				rprintf(FERROR, "write failed on %s: %s\n",
-					full_fname(fname), strerror(errno));
-				exit_cleanup(RERR_FILEIO);
-			}
+			if (fd != -1 && write_file(fd,data,i) != i)
+				goto report_write_error;
 			offset += i;
 			continue;
 		}
@@ -233,7 +253,7 @@ static int receive_data(int f_in,struct map_struct *mapbuf,int fd,char *fname,
 		i = -(i+1);
 		offset2 = i*(OFF_T)sum.blength;
 		len = sum.blength;
-		if (i == (int) sum.count-1 && sum.remainder != 0)
+		if (i == (int)sum.count-1 && sum.remainder != 0)
 			len = sum.remainder;
 
 		stats.matched_data += len;
@@ -249,61 +269,83 @@ static int receive_data(int f_in,struct map_struct *mapbuf,int fd,char *fname,
 			sum_update(map,len);
 		}
 
-		if (fd != -1 && write_file(fd,map,len) != (int) len) {
-			rprintf(FERROR, "write failed on %s: %s\n",
-				full_fname(fname), strerror(errno));
-			exit_cleanup(RERR_FILEIO);
+		if (inplace) {
+			if (offset == offset2 && fd != -1) {
+				if (flush_write_file(fd) < 0)
+					goto report_write_error;
+				offset += len;
+				if (do_lseek(fd, len, SEEK_CUR) != offset) {
+					rsyserr(FERROR, errno,
+						"lseek failed on %s",
+						full_fname(fname));
+					exit_cleanup(RERR_FILEIO);
+				}
+				continue;
+			}
 		}
+		if (fd != -1 && write_file(fd, map, len) != (int)len)
+			goto report_write_error;
 		offset += len;
 	}
 
 	flush_write_file(fd);
 
+#ifdef HAVE_FTRUNCATE
+	if (inplace && fd != -1)
+		ftruncate(fd, offset);
+#endif
+
 	if (do_progress)
 		end_progress(total_size);
 
 	if (fd != -1 && offset > 0 && sparse_end(fd) != 0) {
-		rprintf(FERROR, "write failed on %s: %s\n",
-			full_fname(fname), strerror(errno));
+	    report_write_error:
+		rsyserr(FERROR, errno, "write failed on %s",
+			full_fname(fname));
 		exit_cleanup(RERR_FILEIO);
 	}
 
 	sum_end(file_sum1);
 
+	if (mapbuf)
+		unmap_file(mapbuf);
+
 	read_buf(f_in,file_sum2,MD4_SUM_LENGTH);
-	if (verbose > 2) {
+	if (verbose > 2)
 		rprintf(FINFO,"got file_sum\n");
-	}
-	if (fd != -1 && memcmp(file_sum1,file_sum2,MD4_SUM_LENGTH) != 0) {
+	if (fd != -1 && memcmp(file_sum1, file_sum2, MD4_SUM_LENGTH) != 0)
 		return 0;
-	}
 	return 1;
 }
 
 
+static void discard_receive_data(int f_in, OFF_T length)
+{
+	receive_data(f_in, NULL, -1, 0, NULL, -1, length);
+}
+
+
 /**
  * main routine for receiver process.
  *
  * Receiver process runs on the same host as the generator process. */
-int recv_files(int f_in,struct file_list *flist,char *local_name)
+int recv_files(int f_in, struct file_list *flist, char *local_name)
 {
+	int next_gen_i = -1;
 	int fd1,fd2;
 	STRUCT_STAT st;
 	char *fname, fbuf[MAXPATHLEN];
 	char template[MAXPATHLEN];
 	char fnametmp[MAXPATHLEN];
-	char *fnamecmp;
+	char *fnamecmp, *partialptr;
 	char fnamecmpbuf[MAXPATHLEN];
-	struct map_struct *mapbuf;
-	int i;
 	struct file_struct *file;
-	int phase=0;
-	int recv_ok;
 	struct stats initial_stats;
+	int save_make_backups = make_backups;
+	int i, recv_ok, phase = 0;
 
-	if (verbose > 2) {
+	if (verbose > 2)
 		rprintf(FINFO,"recv_files(%d) starting\n",flist->count);
-	}
 
 	if (flist->hlink_pool) {
 		pool_destroy(flist->hlink_pool);
@@ -315,15 +357,23 @@ int recv_files(int f_in,struct file_list *flist,char *local_name)
 
 		i = read_int(f_in);
 		if (i == -1) {
-			if (phase == 0) {
-				phase++;
-				csum_length = SUM_LENGTH;
-				if (verbose > 2)
-					rprintf(FINFO,"recv_files phase=%d\n",phase);
-				send_msg(MSG_DONE, "", 0);
-				continue;
+			if (read_batch) {
+				if (next_gen_i != flist->count)
+					while (read_int(batch_gen_fd) != -1) {}
+				next_gen_i = -1;
 			}
-			break;
+
+			if (phase)
+				break;
+
+			phase = 1;
+			csum_length = SUM_LENGTH;
+			if (verbose > 2)
+				rprintf(FINFO, "recv_files phase=%d\n", phase);
+			send_msg(MSG_DONE, "", 0);
+			if (keep_partial)
+				make_backups = 0; /* prevents double backup */
+			continue;
 		}
 
 		if (i < 0 || i >= flist->count) {
@@ -345,22 +395,58 @@ int recv_files(int f_in,struct file_list *flist,char *local_name)
 			fname = f_name_to(file, fbuf);
 
 		if (dry_run) {
-			if (!am_server && verbose) {	/* log transfer */
-				rprintf(FINFO, "%s\n", fname);
-			}
+			if (!am_server && verbose) /* log the transfer */
+				rprintf(FINFO, "%s\n", safe_fname(fname));
 			continue;
 		}
 
 		initial_stats = stats;
 
 		if (verbose > 2)
-			rprintf(FINFO,"recv_files(%s)\n",fname);
+			rprintf(FINFO, "recv_files(%s)\n", safe_fname(fname));
 
-		fnamecmp = fname;
+		if (read_batch) {
+			while (i > next_gen_i) {
+				next_gen_i = read_int(batch_gen_fd);
+				if (next_gen_i == -1)
+					next_gen_i = flist->count;
+			}
+			if (i < next_gen_i) {
+				rprintf(FINFO, "skipping update for \"%s\"\n",
+					safe_fname(fname));
+				discard_receive_data(f_in, file->length);
+				continue;
+			}
+		}
+
+		if (server_exclude_list.head
+		    && check_exclude(&server_exclude_list, fname,
+				     S_ISDIR(file->mode)) < 0) {
+			rprintf(FERROR, "attempt to hack rsync failed.\n");
+			exit_cleanup(RERR_PROTOCOL);
+		}
+
+		if (partial_dir) {
+			if ((partialptr = partial_dir_fname(fname)) != NULL)
+				fnamecmp = partialptr;
+			else
+				fnamecmp = fname;
+		} else
+			fnamecmp = partialptr = fname;
+
+		if (inplace && make_backups) {
+			if (!(fnamecmp = get_backup_name(fname)))
+				fnamecmp = partialptr;
+		}
 
 		/* open the file */
 		fd1 = do_open(fnamecmp, O_RDONLY, 0);
 
+		if (fd1 == -1 && fnamecmp != fname) {
+			fnamecmp = fname;
+			fd1 = do_open(fnamecmp, O_RDONLY, 0);
+		}
+
 		if (fd1 == -1 && compare_dest != NULL) {
 			/* try the file at compare_dest instead */
 			pathjoin(fnamecmpbuf, sizeof fnamecmpbuf,
@@ -370,9 +456,9 @@ int recv_files(int f_in,struct file_list *flist,char *local_name)
 		}
 
 		if (fd1 != -1 && do_fstat(fd1,&st) != 0) {
-			rprintf(FERROR, "fstat %s failed: %s\n",
-				full_fname(fnamecmp), strerror(errno));
-			receive_data(f_in,NULL,-1,NULL,file->length);
+			rsyserr(FERROR, errno, "fstat %s failed",
+				full_fname(fnamecmp));
+			discard_receive_data(f_in, file->length);
 			close(fd1);
 			continue;
 		}
@@ -385,7 +471,7 @@ int recv_files(int f_in,struct file_list *flist,char *local_name)
 			 */
 			rprintf(FERROR,"recv_files: %s is a directory\n",
 				full_fname(fnamecmp));
-			receive_data(f_in, NULL, -1, NULL, file->length);
+			discard_receive_data(f_in, file->length);
 			close(fd1);
 			continue;
 		}
@@ -393,7 +479,6 @@ int recv_files(int f_in,struct file_list *flist,char *local_name)
 		if (fd1 != -1 && !S_ISREG(st.st_mode)) {
 			close(fd1);
 			fd1 = -1;
-			mapbuf = NULL;
 		}
 
 		if (fd1 != -1 && !preserve_perms) {
@@ -403,86 +488,123 @@ int recv_files(int f_in,struct file_list *flist,char *local_name)
 			file->mode = st.st_mode;
 		}
 
-		if (fd1 != -1 && st.st_size > 0) {
-			mapbuf = map_file(fd1,st.st_size);
-			if (verbose > 2)
-				rprintf(FINFO,"recv mapped %s of size %.0f\n",fnamecmp,(double)st.st_size);
-		} else
-			mapbuf = NULL;
+		/* We now check to see if we are writing file "inplace" */
+		if (inplace)  {
+			fd2 = do_open(fname, O_WRONLY|O_CREAT, 0);
+			if (fd2 == -1) {
+				rsyserr(FERROR, errno, "open %s failed",
+					full_fname(fname));
+				discard_receive_data(f_in, file->length);
+				if (fd1 != -1)
+					close(fd1);
+				continue;
+			}
+		} else {
+			if (!get_tmpname(fnametmp,fname)) {
+				discard_receive_data(f_in, file->length);
+				if (fd1 != -1)
+					close(fd1);
+				continue;
+			}
 
-		if (!get_tmpname(fnametmp,fname)) {
-			if (mapbuf) unmap_file(mapbuf);
-			if (fd1 != -1) close(fd1);
-			continue;
-		}
+			strlcpy(template, fnametmp, sizeof template);
 
-		strlcpy(template, fnametmp, sizeof template);
-
-		/* we initially set the perms without the
-		 * setuid/setgid bits to ensure that there is no race
-		 * condition. They are then correctly updated after
-		 * the lchown. Thanks to snabb@epipe.fi for pointing
-		 * this out.  We also set it initially without group
-		 * access because of a similar race condition. */
-		fd2 = do_mkstemp(fnametmp, file->mode & INITACCESSPERMS);
-
-		/* in most cases parent directories will already exist
-		 * because their information should have been previously
-		 * transferred, but that may not be the case with -R */
-		if (fd2 == -1 && relative_paths && errno == ENOENT &&
-		    create_directory_path(fnametmp, orig_umask) == 0) {
-			strlcpy(fnametmp, template, sizeof fnametmp);
+			/* we initially set the perms without the
+			 * setuid/setgid bits to ensure that there is no race
+			 * condition. They are then correctly updated after
+			 * the lchown. Thanks to snabb@epipe.fi for pointing
+			 * this out.  We also set it initially without group
+			 * access because of a similar race condition. */
 			fd2 = do_mkstemp(fnametmp, file->mode & INITACCESSPERMS);
-		}
-		if (fd2 == -1) {
-			rprintf(FERROR, "mkstemp %s failed: %s\n",
-				full_fname(fnametmp), strerror(errno));
-			receive_data(f_in,mapbuf,-1,NULL,file->length);
-			if (mapbuf) unmap_file(mapbuf);
-			if (fd1 != -1) close(fd1);
-			continue;
+
+			/* in most cases parent directories will already exist
+			 * because their information should have been previously
+			 * transferred, but that may not be the case with -R */
+			if (fd2 == -1 && relative_paths && errno == ENOENT
+			    && create_directory_path(fnametmp, orig_umask) == 0) {
+				strlcpy(fnametmp, template, sizeof fnametmp);
+				fd2 = do_mkstemp(fnametmp, file->mode & INITACCESSPERMS);
+			}
+			if (fd2 == -1) {
+				rsyserr(FERROR, errno, "mkstemp %s failed",
+					full_fname(fnametmp));
+				discard_receive_data(f_in, file->length);
+				if (fd1 != -1)
+					close(fd1);
+				continue;
+			}
+
+			if (partialptr)
+				cleanup_set(fnametmp, partialptr, file, fd1, fd2);
 		}
 
-		cleanup_set(fnametmp, fname, file, mapbuf, fd1, fd2);
-
-		if (!am_server && verbose) {	/* log transfer */
-			rprintf(FINFO, "%s\n", fname);
-		}
+		if (!am_server && verbose) /* log the transfer */
+			rprintf(FINFO, "%s\n", safe_fname(fname));
 
 		/* recv file data */
-		recv_ok = receive_data(f_in,mapbuf,fd2,fname,file->length);
+		recv_ok = receive_data(f_in, fnamecmp, fd1, st.st_size,
+				       fname, fd2, file->length);
 
 		log_recv(file, &initial_stats);
 
-		if (mapbuf) unmap_file(mapbuf);
-		if (fd1 != -1) {
+		if (fd1 != -1)
 			close(fd1);
+		if (close(fd2) < 0) {
+			rsyserr(FERROR, errno, "close failed on %s",
+				full_fname(fnametmp));
+			exit_cleanup(RERR_FILEIO);
 		}
-		close(fd2);
 
-		if (verbose > 2)
-			rprintf(FINFO,"renaming %s to %s\n",fnametmp,fname);
+		if (recv_ok || inplace)
+			finish_transfer(fname, fnametmp, file, recv_ok);
+		else if (keep_partial && partialptr
+		    && handle_partial_dir(partialptr, PDIR_CREATE))
+			finish_transfer(partialptr, fnametmp, file, 0);
+		else {
+			partialptr = NULL;
+			do_unlink(fnametmp);
+		}
 
-		finish_transfer(fname, fnametmp, file);
+		if (partialptr != fname && fnamecmp == partialptr && recv_ok) {
+			do_unlink(partialptr);
+			handle_partial_dir(partialptr, PDIR_DELETE);
+		}
 
 		cleanup_disable();
 
 		if (!recv_ok) {
-			if (csum_length == SUM_LENGTH) {
-				rprintf(FERROR,"ERROR: file corruption in %s. File changed during transfer?\n",
-					full_fname(fname));
-			} else {
+			int msgtype = csum_length == SUM_LENGTH || read_batch ?
+				FERROR : FINFO;
+			if (msgtype == FERROR || verbose) {
+				char *errstr, *redostr, *keptstr;
+				if (!(keep_partial && partialptr) && !inplace)
+					keptstr = "discarded";
+				else if (partial_dir)
+					keptstr = "put into partial-dir";
+				else
+					keptstr = "retained";
+				if (msgtype == FERROR) {
+					errstr = "ERROR";
+					redostr = "";
+				} else {
+					errstr = "WARNING";
+					redostr = " (will try again)";
+				}
+				rprintf(msgtype,
+					"%s: %s failed verification -- update %s%s.\n",
+					errstr, safe_fname(fname),
+					keptstr, redostr);
+			}
+			if (csum_length != SUM_LENGTH) {
 				char buf[4];
-				if (verbose > 1)
-					rprintf(FINFO,"redoing %s(%d)\n",fname,i);
 				SIVAL(buf, 0, i);
 				send_msg(MSG_REDO, buf, 4);
 			}
 		}
 	}
+	make_backups = save_make_backups;
 
-	if (delete_after && recurse && delete_mode && !local_name
-	    && flist->count > 0)
+	if (delete_after && recurse && !local_name && flist->count > 0)
 		delete_files(flist);
 
 	if (verbose > 2)
