@@ -26,13 +26,16 @@ extern int verbose;
 extern int dry_run;
 extern int preserve_times;
 extern int am_root;
-extern int am_server;
 extern int am_sender;
 extern int am_generator;
 extern int preserve_uid;
 extern int preserve_gid;
-extern int preserve_perms;
+extern int force_delete;
+extern int recurse;
+extern int keep_dirlinks;
 extern int make_backups;
+extern char *backup_dir;
+extern int inplace;
 
 
 /*
@@ -54,10 +57,8 @@ int delete_file(char *fname)
 	DIR *d;
 	struct dirent *di;
 	char buf[MAXPATHLEN];
-	extern int force_delete;
 	STRUCT_STAT st;
 	int ret;
-	extern int recurse;
 
 #if SUPPORT_LINKS
 	ret = do_lstat(fname, &st);
@@ -70,8 +71,8 @@ int delete_file(char *fname)
 	if (!S_ISDIR(st.st_mode)) {
 		if (robust_unlink(fname) == 0 || errno == ENOENT)
 			return 0;
-		rprintf(FERROR, "delete_file: unlink %s failed: %s\n",
-			full_fname(fname), strerror(errno));
+		rsyserr(FERROR, errno, "delete_file: unlink %s failed",
+			full_fname(fname));
 		return -1;
 	}
 
@@ -79,15 +80,15 @@ int delete_file(char *fname)
 		return 0;
 	if (!force_delete || !recurse
 	    || (errno != ENOTEMPTY && errno != EEXIST)) {
-		rprintf(FERROR, "delete_file: rmdir %s failed: %s\n",
-			full_fname(fname), strerror(errno));
+		rsyserr(FERROR, errno, "delete_file: rmdir %s failed",
+			full_fname(fname));
 		return -1;
 	}
 
 	/* now we do a recsursive delete on the directory ... */
 	if (!(d = opendir(fname))) {
-		rprintf(FERROR, "delete_file: opendir %s failed: %s\n",
-			full_fname(fname), strerror(errno));
+		rsyserr(FERROR, errno, "delete_file: opendir %s failed",
+			full_fname(fname));
 		return -1;
 	}
 
@@ -98,15 +99,15 @@ int delete_file(char *fname)
 			continue;
 		pathjoin(buf, sizeof buf, fname, dname);
 		if (verbose > 0)
-			rprintf(FINFO, "deleting %s\n", buf);
+			rprintf(FINFO, "deleting %s\n", safe_fname(buf));
 		if (delete_file(buf) != 0) {
 			closedir(d);
 			return -1;
 		}
 	}
 	if (errno) {
-		rprintf(FERROR, "delete_file: readdir %s failed: %s\n",
-			full_fname(fname), strerror(errno));
+		rsyserr(FERROR, errno, "delete_file: readdir %s failed",
+			full_fname(fname));
 		closedir(d);
 		return -1;
 	}
@@ -114,8 +115,8 @@ int delete_file(char *fname)
 	closedir(d);
 
 	if (do_rmdir(fname) != 0) {
-		rprintf(FERROR, "delete_file: rmdir %s failed: %s\n",
-			full_fname(fname), strerror(errno));
+		rsyserr(FERROR, errno, "delete_file: rmdir %s failed",
+			full_fname(fname));
 		return -1;
 	}
 
@@ -123,31 +124,35 @@ int delete_file(char *fname)
 }
 
 int set_perms(char *fname,struct file_struct *file,STRUCT_STAT *st,
-		int report)
+	      int flags)
 {
 	int updated = 0;
 	STRUCT_STAT st2;
 	int change_uid, change_gid;
 
-	if (dry_run) return 0;
+	if (dry_run)
+		return 0;
 
 	if (!st) {
-		if (link_stat(fname,&st2) != 0) {
-			rprintf(FERROR, "stat %s failed: %s\n",
-				full_fname(fname), strerror(errno));
+		if (link_stat(fname, &st2, 0) < 0) {
+			rsyserr(FERROR, errno, "stat %s failed",
+				full_fname(fname));
 			return 0;
 		}
 		st = &st2;
 	}
 
-	if (preserve_times && !S_ISLNK(st->st_mode) &&
-	    cmp_modtime(st->st_mtime, file->modtime) != 0) {
+	if (!preserve_times || S_ISLNK(st->st_mode)
+	    || (make_backups && !backup_dir && S_ISDIR(st->st_mode)))
+		flags |= PERMS_SKIP_MTIME;
+	if (!(flags & PERMS_SKIP_MTIME)
+	    && cmp_modtime(st->st_mtime, file->modtime) != 0) {
 		/* don't complain about not setting times on directories
 		 * because some filesystems can't do it */
 		if (set_modtime(fname,file->modtime) != 0 &&
 		    !S_ISDIR(st->st_mode)) {
-			rprintf(FERROR, "failed to set times on %s: %s\n",
-				full_fname(fname), strerror(errno));
+			rsyserr(FERROR, errno, "failed to set times on %s",
+				full_fname(fname));
 			return 0;
 		}
 		updated = 1;
@@ -174,16 +179,17 @@ int set_perms(char *fname,struct file_struct *file,STRUCT_STAT *st,
 		    change_gid ? file->gid : st->st_gid) != 0) {
 			/* shouldn't have attempted to change uid or gid
 			 * unless have the privilege */
-			rprintf(FERROR, "%s %s failed: %s\n",
+			rsyserr(FERROR, errno, "%s %s failed",
 			    change_uid ? "chown" : "chgrp",
-			    full_fname(fname), strerror(errno));
+			    full_fname(fname));
 			return 0;
 		}
 		/* a lchown had been done - we have to re-stat if the
                  * destination had the setuid or setgid bits set due
                  * to the side effect of the chown call */
 		if (st->st_mode & (S_ISUID | S_ISGID)) {
-			link_stat(fname, st);
+			link_stat(fname, st,
+				  keep_dirlinks && S_ISDIR(st->st_mode));
 		}
 		updated = 1;
 	}
@@ -193,15 +199,15 @@ int set_perms(char *fname,struct file_struct *file,STRUCT_STAT *st,
 		if ((st->st_mode & CHMOD_BITS) != (file->mode & CHMOD_BITS)) {
 			updated = 1;
 			if (do_chmod(fname,(file->mode & CHMOD_BITS)) != 0) {
-				rprintf(FERROR, "failed to set permissions on %s: %s\n",
-					full_fname(fname), strerror(errno));
+				rsyserr(FERROR, errno, "failed to set permissions on %s",
+					full_fname(fname));
 				return 0;
 			}
 		}
 	}
 #endif
 
-	if (verbose > 1 && report) {
+	if (verbose > 1 && flags & PERMS_REPORT) {
 		if (updated)
 			rprintf(FINFO,"%s\n",fname);
 		else
@@ -228,26 +234,43 @@ void sig_int(void)
 
 /* finish off a file transfer, renaming the file and setting the permissions
    and ownership */
-void finish_transfer(char *fname, char *fnametmp, struct file_struct *file)
+void finish_transfer(char *fname, char *fnametmp, struct file_struct *file,
+		     int ok_to_set_time)
 {
 	int ret;
 
+	if (inplace) {
+		if (verbose > 2)
+			rprintf(FINFO, "finishing %s\n", fname);
+		goto do_set_perms;
+	}
+
 	if (make_backups && !make_backup(fname))
 		return;
 
+	/* Change permissions before putting the file into place. */
+	set_perms(fnametmp, file, NULL, ok_to_set_time ? 0 : PERMS_SKIP_MTIME);
+
 	/* move tmp file over real file */
+	if (verbose > 2)
+		rprintf(FINFO, "renaming %s to %s\n", fnametmp, fname);
 	ret = robust_rename(fnametmp, fname, file->mode & INITACCESSPERMS);
 	if (ret < 0) {
-		rprintf(FERROR, "%s %s -> \"%s\": %s\n",
+		rsyserr(FERROR, errno, "%s %s -> \"%s\"",
 		    ret == -2 ? "copy" : "rename",
-		    full_fname(fnametmp), fname, strerror(errno));
+		    full_fname(fnametmp), fname);
 		do_unlink(fnametmp);
-	} else {
-		set_perms(fname,file,NULL,0);
+		return;
 	}
+	if (ret == 0) {
+		/* The file was moved into place (not copied), so it's done. */
+		return;
+	}
+    do_set_perms:
+	set_perms(fname, file, NULL, ok_to_set_time ? 0 : PERMS_SKIP_MTIME);
 }
 
 const char *who_am_i(void)
 {
-    return am_sender ? "sender" : am_generator ? "generator" : "receiver";
+	return am_sender ? "sender" : am_generator ? "generator" : "receiver";
 }
