@@ -1,17 +1,17 @@
-/* 
+/*
    Copyright (C) Andrew Tridgell 1996
    Copyright (C) Paul Mackerras 1996
-   
+
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
-   
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-   
+
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
@@ -21,23 +21,26 @@
 #include "zlib/zlib.h"
 
 extern int do_compression;
+extern int module_id;
+
 static int compression_level = Z_DEFAULT_COMPRESSION;
 
 /* determine the compression level based on a wildcard filename list */
 void set_compression(char *fname)
 {
-	extern int module_id;
 	char *dont;
 	char *tok;
 
-	if (!do_compression) return;
+	if (!do_compression)
+		return;
 
 	compression_level = Z_DEFAULT_COMPRESSION;
 	dont = lp_dont_compress(module_id);
 
-	if (!dont || !*dont) return;
+	if (!dont || !*dont)
+		return;
 
-	if ((dont[0] == '*') && (!dont[1])) {
+	if (dont[0] == '*' && !dont[1]) {
 		/* an optimization to skip the rest of this routine */
 		compression_level = 0;
 		return;
@@ -45,13 +48,14 @@ void set_compression(char *fname)
 
 	dont = strdup(dont);
 	fname = strdup(fname);
-	if (!dont || !fname) return;
+	if (!dont || !fname)
+		return;
 
 	strlower(dont);
 	strlower(fname);
 
-	for (tok=strtok(dont," ");tok;tok=strtok(NULL," ")) {
-		if (fnmatch(tok, fname, 0) == 0) {
+	for (tok = strtok(dont, " "); tok; tok = strtok(NULL, " ")) {
+		if (wildmatch(tok, fname)) {
 			compression_level = 0;
 			break;
 		}
@@ -69,12 +73,14 @@ static int simple_recv_token(int f,char **data)
 
 	if (!buf) {
 		buf = new_array(char, CHUNK_SIZE);
-		if (!buf) out_of_memory("simple_recv_token");
+		if (!buf)
+			out_of_memory("simple_recv_token");
 	}
 
 	if (residue == 0) {
 		int i = read_int(f);
-		if (i <= 0) return i;
+		if (i <= 0)
+			return i;
 		residue = i;
 	}
 
@@ -90,29 +96,18 @@ static int simple_recv_token(int f,char **data)
 static void simple_send_token(int f,int token,
 			      struct map_struct *buf,OFF_T offset,int n)
 {
-	extern int write_batch; /* dw */
-	int hold_int; /* dw */
-
 	if (n > 0) {
 		int l = 0;
 		while (l < n) {
 			int n1 = MIN(CHUNK_SIZE,n-l);
 			write_int(f,n1);
 			write_buf(f,map_ptr(buf,offset+l,n1),n1);
-			if (write_batch) {
-			    write_batch_delta_file( (char *) &n1, sizeof(int) );
-			    write_batch_delta_file(map_ptr(buf,offset+l,n1),n1);
-			}
 			l += n1;
 		}
 	}
 	/* a -2 token means to send data only and no token */
 	if (token != -2) {
 		write_int(f,-(token+1));
-		if (write_batch) {
-		    hold_int = -(token+1);
-		    write_batch_delta_file( (char *) &hold_int, sizeof(int) );
-		}
 	}
 }
 
@@ -127,6 +122,12 @@ static void simple_send_token(int f,int token,
 
 #define MAX_DATA_COUNT	16383	/* fit 14 bit count into 2 bytes with flags */
 
+/* zlib.h says that if we want to be able to compress something in a single
+ * call, avail_out must be at least 0.1% larger than avail_in plus 12 bytes.
+ * We'll add in 0.1%+16, just to be safe (and we'll avoid floating point,
+ * to ensure that this is a compile-time value). */
+#define AVAIL_OUT_SIZE(avail_in_size) ((avail_in_size)*1001/1000+16)
+
 /* For coding runs of tokens */
 static int last_token = -1;
 static int run_start;
@@ -138,6 +139,14 @@ static z_stream tx_strm;
 /* Output buffer */
 static char *obuf;
 
+/* We want obuf to be able to hold both MAX_DATA_COUNT+2 bytes as well as
+ * AVAIL_OUT_SIZE(CHUNK_SIZE) bytes, so make sure that it's large enough. */
+#if MAX_DATA_COUNT+2 > AVAIL_OUT_SIZE(CHUNK_SIZE)
+#define OBUF_SIZE	(MAX_DATA_COUNT+2)
+#else
+#define OBUF_SIZE	AVAIL_OUT_SIZE(CHUNK_SIZE)
+#endif
+
 /* Send a deflated token */
 static void
 send_deflated_token(int f, int token,
@@ -145,8 +154,6 @@ send_deflated_token(int f, int token,
 {
 	int n, r;
 	static int init_done, flush_pending;
-	extern int write_batch;  /* dw */
-	char temp_byte;   /* dw */
 
 	if (last_token == -1) {
 		/* initialization */
@@ -160,7 +167,7 @@ send_deflated_token(int f, int token,
 				rprintf(FERROR, "compression init failed\n");
 				exit_cleanup(RERR_STREAMIO);
 			}
-			if ((obuf = new_array(char, MAX_DATA_COUNT+2)) == NULL)
+			if ((obuf = new_array(char, OBUF_SIZE)) == NULL)
 				out_of_memory("send_deflated_token");
 			init_done = 1;
 		} else
@@ -179,27 +186,13 @@ send_deflated_token(int f, int token,
 		n = last_token - run_start;
 		if (r >= 0 && r <= 63) {
 			write_byte(f, (n==0? TOKEN_REL: TOKENRUN_REL) + r);
-			if (write_batch) { /* dw */
-			    temp_byte = (char)( (n==0? TOKEN_REL: TOKENRUN_REL) + r);
-			    write_batch_delta_file(&temp_byte,sizeof(char));
-			}
 		} else {
 			write_byte(f, (n==0? TOKEN_LONG: TOKENRUN_LONG));
 			write_int(f, run_start);
-			if (write_batch) { /* dw */
-			    temp_byte = (char)(n==0? TOKEN_LONG: TOKENRUN_LONG);
-			    write_batch_delta_file(&temp_byte,sizeof(temp_byte));
-			    write_batch_delta_file((char *)&run_start,sizeof(run_start));
-			}
 		}
 		if (n != 0) {
 			write_byte(f, n);
 			write_byte(f, n >> 8);
-			if (write_batch) { /* dw */
-			    write_batch_delta_file((char *)&n,sizeof(char));
-			    temp_byte = (char) n >> 8;
-			    write_batch_delta_file(&temp_byte,sizeof(temp_byte));
-			}
 		}
 		last_run_end = last_token;
 		run_start = token;
@@ -258,8 +251,6 @@ send_deflated_token(int f, int token,
 					obuf[0] = DEFLATED_DATA + (n >> 8);
 					obuf[1] = n;
 					write_buf(f, obuf, n+2);
-					if (write_batch) /* dw */
-					    write_batch_delta_file(obuf,n+2);
 				}
 			}
 		} while (nb != 0 || tx_strm.avail_out == 0);
@@ -269,18 +260,13 @@ send_deflated_token(int f, int token,
 	if (token == -1) {
 		/* end of file - clean up */
 		write_byte(f, END_FLAG);
-		if (write_batch) { /* dw */
-		    temp_byte = END_FLAG;
-		    write_batch_delta_file((char *)&temp_byte,sizeof(temp_byte));
-		}
-
 	} else if (token != -2) {
 		/* add the data in the current block to the compressor's
 		   history and hash table */
 		tx_strm.next_in = (Bytef *) map_ptr(buf, offset, toklen);
 		tx_strm.avail_in = toklen;
 		tx_strm.next_out = (Bytef *) obuf;
-		tx_strm.avail_out = MAX_DATA_COUNT;
+		tx_strm.avail_out = AVAIL_OUT_SIZE(CHUNK_SIZE);
 		r = deflate(&tx_strm, Z_INSERT_ONLY);
 		if (r != Z_OK || tx_strm.avail_in != 0) {
 			rprintf(FERROR, "deflate on token returned %d (%d bytes left)\n",
@@ -323,7 +309,7 @@ recv_deflated_token(int f, char **data)
 					exit_cleanup(RERR_STREAMIO);
 				}
 				if (!(cbuf = new_array(char, MAX_DATA_COUNT))
-				    || !(dbuf = new_array(char, CHUNK_SIZE)))
+				    || !(dbuf = new_array(char, AVAIL_OUT_SIZE(CHUNK_SIZE))))
 					out_of_memory("recv_deflated_token");
 				init_done = 1;
 			} else {
@@ -352,9 +338,9 @@ recv_deflated_token(int f, char **data)
 				/* check previous inflated stuff ended correctly */
 				rx_strm.avail_in = 0;
 				rx_strm.next_out = (Bytef *)dbuf;
-				rx_strm.avail_out = CHUNK_SIZE;
+				rx_strm.avail_out = AVAIL_OUT_SIZE(CHUNK_SIZE);
 				r = inflate(&rx_strm, Z_SYNC_FLUSH);
-				n = CHUNK_SIZE - rx_strm.avail_out;
+				n = AVAIL_OUT_SIZE(CHUNK_SIZE) - rx_strm.avail_out;
 				/*
 				 * Z_BUF_ERROR just means no progress was
 				 * made, i.e. the decompressor didn't have
@@ -408,9 +394,9 @@ recv_deflated_token(int f, char **data)
 
 		case r_inflating:
 			rx_strm.next_out = (Bytef *)dbuf;
-			rx_strm.avail_out = CHUNK_SIZE;
+			rx_strm.avail_out = AVAIL_OUT_SIZE(CHUNK_SIZE);
 			r = inflate(&rx_strm, Z_NO_FLUSH);
-			n = CHUNK_SIZE - rx_strm.avail_out;
+			n = AVAIL_OUT_SIZE(CHUNK_SIZE) - rx_strm.avail_out;
 			if (r != Z_OK) {
 				rprintf(FERROR, "inflate returned %d (%d bytes)\n", r, n);
 				exit_cleanup(RERR_STREAMIO);
@@ -465,7 +451,7 @@ static void see_deflate_token(char *buf, int len)
 			}
 		}
 		rx_strm.next_out = (Bytef *)dbuf;
-		rx_strm.avail_out = CHUNK_SIZE;
+		rx_strm.avail_out = AVAIL_OUT_SIZE(CHUNK_SIZE);
 		r = inflate(&rx_strm, Z_SYNC_FLUSH);
 		if (r != Z_OK) {
 			rprintf(FERROR, "inflate (token) returned %d\n", r);
@@ -476,7 +462,7 @@ static void see_deflate_token(char *buf, int len)
 
 /**
  * Transmit a verbatim buffer of length @p n followed by a token.
- * If token == -1 then we have reached EOF 
+ * If token == -1 then we have reached EOF
  * If n == 0 then don't send a buffer
  */
 void send_token(int f,int token,struct map_struct *buf,OFF_T offset,
